"""Selenium web scraping module."""
from __future__ import annotations

import logging
import asyncio
from pathlib import Path
from sys import platform

from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.safari.options import Options as SafariOptions
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from fastapi import WebSocket

from scraping import scrape_skills, processing as summary
from scraping.processing.html import extract_hyperlinks, format_hyperlinks

from concurrent.futures import ThreadPoolExecutor

from scraping.processing.text import summarize_text

executor = ThreadPoolExecutor()

FILE_DIR = Path(__file__).parent.parent


async def async_browse(
        selenium_web_browser: str,
        user_agent: str,
        fast_llm_model: str,
        summary_token_limit: str,
        llm_provider: str,
        url: str, question: str,
        websocket: WebSocket
) -> str:
    """Browse a website and return the answer and links to the user

    Args:
        selenium_web_browser (str): The web browser used for scraping
        user_agent (str): The user agent used when scraping
        url (str): The url of the website to browse
        question (str): The question asked by the user
        websocket (WebSocketManager): The websocket manager

    Returns:
        str: The answer and links to the user
    """
    loop = asyncio.get_event_loop()
    executor = ThreadPoolExecutor(max_workers=8)

    print(f"Scraping url {url} with question {question}")
    if websocket:
        await websocket.send_json(
            {
                "type": "logs",
                "output": f"🔎 Browsing the {url} for relevant about: {question}...",
            }
        )
    else:
        print(f"🔎 Browsing the {url} for relevant about: {question}...")

    try:
        driver, text = await loop.run_in_executor(
            executor, scrape_text_with_selenium, selenium_web_browser, user_agent, url
        )
        await loop.run_in_executor(executor, add_header, driver)
        summary_text = await loop.run_in_executor(
            executor, summarize_text, fast_llm_model, summary_token_limit, llm_provider, url, text, question, driver
        )
        if websocket:
            await websocket.send_json(
                {
                    "type": "logs",
                    "output": f"📝 Information gathered from url {url}: {summary_text}",
                }
            )
        else:
            print(f"📝 Information gathered from url {url}: {summary_text}")

        return f"Information gathered from url {url}: {summary_text}"
    except Exception as e:
        print(f"An error occurred while processing the url {url}: {e}")
        return f"Error processing the url {url}: {e}"


def browse_website(url: str, question: str) -> tuple[str, WebDriver]:
    """Browse a website and return the answer and links to the user

    Args:
        url (str): The url of the website to browse
        question (str): The question asked by the user

    Returns:
        Tuple[str, WebDriver]: The answer and links to the user and the webdriver
    """

    if not url:
        return "A URL was not specified, cancelling request to browse website.", None

    driver, text = scrape_text_with_selenium(url)
    add_header(driver)
    summary_text = summary.summarize_text(url, text, question, driver)

    links = scrape_links_with_selenium(driver, url)

    # Limit links to 5
    if len(links) > 5:
        links = links[:5]

    # write_to_file('research-{0}.txt'.format(url), summary_text + "\nSource Links: {0}\n\n".format(links))

    close_browser(driver)
    return f"Answer gathered from website: {summary_text} \n \n Links: {links}", driver


def scrape_text_with_selenium(selenium_web_browser: str, user_agent: str, url: str) -> tuple[WebDriver, str]:
    """Scrape text from a website using selenium

    Args:
        url (str): The url of the website to scrape
        selenium_web_browser (str): The web browser used to scrape
        user_agent (str): The user agent used when scraping

    Returns:
        Tuple[WebDriver, str]: The webdriver and the text scraped from the website
    """
    logging.getLogger("selenium").setLevel(logging.CRITICAL)

    options_available = {
        "chrome": ChromeOptions,
        "safari": SafariOptions,
        "firefox": FirefoxOptions,
    }

    options = options_available[selenium_web_browser]()
    options.add_argument(f"user-agent={user_agent}")
    options.add_argument("--headless")
    options.add_argument("--enable-javascript")

    if selenium_web_browser == "firefox":
        driver = webdriver.Firefox(options=options)
    elif selenium_web_browser == "safari":
        # Requires a bit more setup on the users end
        # See https://developer.apple.com/documentation/webkit/testing_with_webdriver_in_safari
        driver = webdriver.Safari(options=options)
    else:
        if platform == "linux" or platform == "linux2":
            options.add_argument("--disable-dev-shm-usage")
            options.add_argument("--remote-debugging-port=9222")
        options.add_argument("--no-sandbox")
        options.add_experimental_option("prefs", {"download_restrictions": 3})
        driver = webdriver.Chrome(options=options)

    print(f"scraping url {url}...")
    driver.get(url)

    WebDriverWait(driver, 10).until(
        EC.presence_of_element_located((By.TAG_NAME, "body"))
    )

    # check if url is a pdf or arxiv link
    if url.endswith(".pdf"):
        text = scrape_skills.scrape_pdf_with_pymupdf(url)
    elif "arxiv" in url:
        # parse the document number from the url
        doc_num = url.split("/")[-1]
        text = scrape_skills.scrape_pdf_with_arxiv(doc_num)
    else:
        # Get the HTML content directly from the browser's DOM
        page_source = driver.execute_script("return document.body.outerHTML;")
        soup = BeautifulSoup(page_source, "html.parser")

        for script in soup(["script", "style"]):
            script.extract()

        # text = soup.get_text()
        text = get_text(soup)

    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    text = "\n".join(chunk for chunk in chunks if chunk)
    return driver, text


def get_text(soup):
    """Get the text from the soup

    Args:
        soup (BeautifulSoup): The soup to get the text from

    Returns:
        str: The text from the soup
    """
    text = ""
    tags = ["h1", "h2", "h3", "h4", "h5", "p"]
    for element in soup.find_all(tags):  # Find all the <p> elements
        text += element.text + "\n\n"
    return text


def scrape_links_with_selenium(driver: WebDriver, url: str) -> list[str]:
    """Scrape links from a website using selenium

    Args:
        driver (WebDriver): The webdriver to use to scrape the links

    Returns:
        List[str]: The links scraped from the website
    """
    page_source = driver.page_source
    soup = BeautifulSoup(page_source, "html.parser")

    for script in soup(["script", "style"]):
        script.extract()

    hyperlinks = extract_hyperlinks(soup, url)

    return format_hyperlinks(hyperlinks)


def close_browser(driver: WebDriver) -> None:
    """Close the browser

    Args:
        driver (WebDriver): The webdriver to close

    Returns:
        None
    """
    driver.quit()


def add_header(driver: WebDriver) -> None:
    """Add a header to the website

    Args:
        driver (WebDriver): The webdriver to use to add the header

    Returns:
        None
    """
    driver.execute_script(open(f"{FILE_DIR}/js/overlay.js", "r").read())



from langchain_core.messages import ToolMessage

from mtmai.agents.states import MtmState


# This node will be shared for exiting all specialized assistants
def pop_dialog_state(state: MtmState) -> dict:
    """Pop the dialog stack and return to the main assistant.

    This lets the full graph explicitly track the dialog flow and delegate control
    to specific sub-graphs.
    """
    messages = []
    if state.messages[-1].tool_calls:
        # Note: Doesn't currently handle the edge case where the llm performs parallel tool calls
        messages.append(
            ToolMessage(
                content="Resuming dialog with the host assistant. Please reflect on the past conversation and assist the user as needed.",
                tool_call_id=state.messages[-1].tool_calls[0]["id"],
            )
        )
    return {
        "dialog_state": "pop",
        "messages": messages,
    }



import argparse
import logging
import sys
from pathlib import Path

from dotenv import load_dotenv

from mtmtrain.core.config import settings

sys.path.insert(0, str(Path("mtmlib").resolve()))
from mtmlib.logging import LoggingOptions, setup_logging

load_dotenv()

setup_logging(option=LoggingOptions())
logger = logging.getLogger("mtmtrain")


"""
    在coleb 的使用方式:
    !pip install -U --no-cache-dir "mtmtrain" --index-url https://pypi.org/simple \
    && mtmtrain init
"""


def task_a(alpha):
    print("task a", alpha)


def task_b(beta, gamma):
    print("task b", beta, gamma)


def worker():
    if not settings.MTMAI_API_BASE:
        print("require server url")
        return
    from mtmtrain.worker import worker_start

    worker_start()


def down_dataset():
    if not settings.MTMAI_API_BASE:
        print("require server url")
        return
    from mtmai_client.api.dataset import dataset_dataset_download
    from mtmai_client.client import Client

    client = Client(base_url=settings.MTMAI_API_BASE)

    local_dataset_dir = ".vol/datasets"
    all_datasets = ["text_classify/bbc-text.csv"]
    for ds in all_datasets:
        response = dataset_dataset_download.sync(client=client, dataset_path=ds)
        dataset_path = Path(local_dataset_dir).joinpath(ds)
        dataset_path.parent.mkdir(parents=True, exist_ok=True)
        dataset_path.write_text(response)
        logger.info("数据集下载完成 %s", dataset_path.resolve())


def main():
    print("[🚀 mtmtrain]")

    parser = argparse.ArgumentParser(description="mtmtrain")

    # 添加全局参数
    parser.add_argument(
        "-s",
        "--server-url",
        help="Specify the backend server URL.",
        default="http://localhost:8444",  # 默认 URL
        type=str,
    )
    subparsers = parser.add_subparsers(dest="subparser")
    parser_a = subparsers.add_parser("task_a")
    parser_a.add_argument("-a", "--alpha", dest="alpha", help="Alpha description")

    parser_b = subparsers.add_parser("task_b")
    parser_b.add_argument("-b", "--beta", dest="beta", help="Beta description")
    parser_b.add_argument(
        "-g", "--gamma", dest="gamma", default=42, help="Gamma description"
    )

    parser_worker = subparsers.add_parser("worker")

    parser_down_dataset = subparsers.add_parser("down_dataset")

    args = parser.parse_args()
    # 设置全局的后端 URL
    settings.MTMAI_API_BASE = args.server_url

    # 调用相应的子命令
    if args.subparser == "task_a":
        task_a(args.alpha)
    elif args.subparser == "task_b":
        task_b(args.beta, args.gamma)
    elif args.subparser == "worker":
        worker()
    elif args.subparser == "down_dataset":
        down_dataset()
    else:
        parser.print_help()

    # kwargs = vars(parser.parse_args())
    # globals()[kwargs.pop("subparser")](**kwargs)

    # # 创建子命令解析器
    # subparsers = parser.add_subparsers(dest="command", help="Subcommand to run")

    # # 定义 'init' 子命令
    # subparsers.add_parser("init", help="Run initialization.")

    # # 定义 'worker' 子命令
    # subparsers.add_parser("worker", help="Start the worker process.")

    # # 定义 'show1' 子命令
    # subparsers.add_parser("show1", help="Show visual output.")
    # subparsers.add_parser("down_dataset", help="Show visual output.")

    # # 定义 'text_classify' 子命令
    # subparsers.add_parser("text_classify", help="Run text classification training.")

    # args = parser.parse_args()

    # if args.command == "init":
    #     from mtmtrain import env

    #     print("Initializing environment...")
    #     env.init_env()
    #     print("Environment initialized.")
    # elif args.command == "worker":
    #     from mtmtrain.worker import worker_start

    #     settings.MTMAI_API_BASE = args.server_url
    #     worker_start()

    # elif args.command == "show1":
    #     from IPython.display import HTML, display

    #     def show_visual_output():
    #         display(HTML("<h1>Processed Data:</h1><pre>88888</pre>"))

    #     show_visual_output()

    # elif args.command == "text_classify":
    #     from mtmtrain import text_classify

    #     text_classify.train()

    # elif args.command == "down_dataset":
    #     from mtmai_client.client import Client

    #     client = Client()
    #     from mtmai_client.api.dataset import dataset_dataset_download

    #     dataset_dataset_download.sync(
    #         client=client, dataset_path="text_classify/bbc-text.csv"
    #     )

    # else:
    #     print("Unknown command", args.command)


if __name__ == "__main__":
    main()



import json
import logging
import re

import fastapi

from mtmai.core.config import settings

router = fastapi.APIRouter()
logger = logging.getLogger()


def register_api_router(app: fastapi.FastAPI):
    app.include_router(router)


@router.get(
    settings.API_V1_STR + "/tools/together_demo",
)
async def tools_together_hello():
    # 学习： 根据 together ai 官方文档 llama3.1 模型不直接支持 tools 调用，
    #       官方使用提示词，并解释content 输出的内容的方式间接实现 函数调用。
    from together import Together

    together = Together(
        api_key="b135fd4bed9be2a988e0376d1bb0977fcb8b6a88ec9f35da8138fa49eb9a0d50"
    )
    weatherTool = {
        "name": "get_current_weather",
        "description": "Get the current weather in a given location",
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. San Francisco, CA",
                },
            },
            "required": ["location"],
        },
    }

    toolPrompt = f"""
    You have access to the following functions:

    Use the function '{weatherTool["name"]}' to '{weatherTool["description"]}':
    {json.dumps(weatherTool)}

    If you choose to call a function ONLY reply in the following format with no prefix or suffix:

    <function=example_function_name>{{\"example_name\": \"example_value\"}}</function>

    Reminder:
    - Function calls MUST follow the specified format, start with <function= and end with </function>
    - Required parameters MUST be specified
    - Only call one function at a time
    - Put the entire function call reply on one line
    - If there is no function call available, answer the question like normal with your current knowledge and do not tell the user about function calls

    """

    messages = [
        {
            "role": "system",
            "content": toolPrompt,
        },
        {
            "role": "user",
            "content": "What is the weather in Tokyo?",
        },
    ]

    response = together.chat.completions.create(
        model="meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
        messages=messages,
        max_tokens=1024,
        temperature=0,
    )

    messages.append(response.choices[0].message)

    def parse_tool_response(response: str):
        function_regex = r"<function=(\w+)>(.*?)</function>"
        match = re.search(function_regex, response)

        if match:
            function_name, args_string = match.groups()
            try:
                args = json.loads(args_string)
                return {  # noqa: TRY300
                    "function": function_name,
                    "arguments": args,
                }
            except json.JSONDecodeError as error:
                print(f"Error parsing function arguments: {error}")
                return None
        return None

    # 解释 llm 的输出
    parsed_response = parse_tool_response(response.choices[0].message.content)
    print(parse_tool_response(response.choices[0].message.content))
    # 输出： {
    # "function": "get_current_weather",
    # "arguments": { "location": "Tokyo, Japan" },
    # }

    return messages



# from langchain_community.agent_toolkits import GmailToolkit
# from langchain_community.tools.gmail.create_draft import GmailCreateDraft
from langchain.tools import tool


class CreateDraftTool:
    @tool("Create Draft")
    def create_draft(data):
        """
        Useful to create an email draft.
        The input to this tool should be a pipe (|) separated text
        of length 3 (three), representing who to send the email to,
        the subject of the email and the actual message.
        For example, `lorem@ipsum.com|Nice To Meet You|Hey it was great to meet you.`.
        """
        email, subject, message = data.split("|")
        # gmail = GmailToolkit()
        # draft = GmailCreateDraft(api_resource=gmail.api_resource)
        # resutl = draft({
        # 		'to': [email],
        # 		'subject': subject,
        # 		'message': message
        # })
        resutl = {
            "to": "fake@email.com",
            "subject": "fakesubject",
            "message": "fake_message",
        }
        return f"\nDraft created: {resutl}\n"



from prefect import get_run_logger, task

from mtmai.models.book_gen import (
    GenBookState,
)


@task()
async def publish_article(
    site_id: str,
    state: GenBookState,
):
    """
    本地博客发布
    """

    logger = get_run_logger()
    logger.info("本地博客发布")
    book_content = ""
    for chapter in state.book:
        # Add the chapter title as an H1 heading
        book_content += f"# {chapter.title}\n\n"
        # Add the chapter content
        book_content += f"{chapter.content}\n\n"

    # The title of the book from self.state.title
    book_title = state.title

    logger.info("文章生成完成, 开始发布")
    filename = (
        f"./{book_title.replace(' ', '_').replace("'", '_').replace('"', '_')}.md"
    )

    logger.info("Book saved as %s", filename)



import jwt
from pydantic import ValidationError

from mtmai.core import security
from mtmai.core.config import settings
from mtmai.models.models import TokenPayload


class TokenDecodeError(Exception):
    def __init__(self, detail: str):
        self.detail = detail


def decode_token(token: str) -> TokenPayload:
    try:
        payload = jwt.decode(
            token, settings.SECRET_KEY, algorithms=[security.ALGORITHM]
        )
        return TokenPayload(**payload)
    except jwt.InvalidTokenError:
        raise TokenDecodeError("Invalid token")
    except jwt.ExpiredSignatureError:
        raise TokenDecodeError("Token has expired")
    except jwt.DecodeError:
        raise TokenDecodeError("Could not decode token")
    except ValidationError:
        raise TokenDecodeError("Invalid token payload")



from collections.abc import Iterable
from typing import TYPE_CHECKING

from fastapi.encoders import jsonable_encoder
from langgraph.graph import StateGraph
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from sqlmodel import Session

from mtmai.models.models import User
from mtmai.mtlibs import aisdk

from .nodes import Nodes
from .state import EmailsState

if TYPE_CHECKING:
    from langchain_core.runnables.config import RunnableConfig


class WorkFlowCrewAgent:
    def __init__(self):
        nodes = Nodes()
        workflow = StateGraph(EmailsState)

        workflow.add_node("check_new_emails", nodes.check_email)
        # workflow.add_node("wait_next_run", nodes.wait_next_run)
        # workflow.add_node("draft_responses", EmailFilterCrew().kickoff)

        workflow.set_entry_point("check_new_emails")
        # workflow.add_conditional_edges(
        #     "check_new_emails",
        #     nodes.new_emails,
        #     {"continue": "draft_responses", "end": "wait_next_run"},
        # )
        # workflow.add_edge("draft_responses", "wait_next_run")
        # workflow.add_edge("wait_next_run", "check_new_emails")
        self.app = workflow.compile()

    @property
    def name(self):
        return "graphcrewdemo"

    async def chat(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        db: Session,
        conversation_id: str | None = None,
        user: User | None = None,
    ):
        wf = self.app
        thread_id = "1"
        input = EmailsState(
            # messages=messages
            checked_emails_ids=["emailid123", "emailid234"]
        )
        config: RunnableConfig = {"configurable": {"thread_id": thread_id}}
        async for event in wf.astream_events(
            input=input,
            version="v2",
            config=config,
        ):
            kind = event["event"]
            name = event["name"]
            data = event["data"]
            if kind == "on_chat_model_stream":
                print("------")
                print(event["data"]["chunk"].dict())
                content = event["data"]["chunk"].content
                if content:
                    yield aisdk.text(content)
                # yield aisdk.text(content)
            print(f"astream_event: kind: {kind}, name={name},{data}")

            if kind == "on_chain_end" and name == "LangGraph":
                # 完全结束可以拿到最终数据
                # yield f"2: {json.dumps(jsonable_encoder(data))}\n"
                yield aisdk.data(jsonable_encoder(data))

        yield aisdk.finish()



"""
convert_to_md module
"""

from urllib.parse import urlparse

import html2text


def convert_to_md(html: str, url: str = None) -> str:
    """Convert HTML to Markdown.
    This function uses the html2text library to convert the provided HTML content to Markdown
    format.
    The function returns the converted Markdown content as a string.

    Args: html (str): The HTML content to be converted.

    Returns: str: The equivalent Markdown content.

    Example: >>> convert_to_md("<html><body><p>This is a paragraph.</p>
    <h1>This is a heading.</h1></body></html>")
    'This is a paragraph.\n\n# This is a heading.'

    Note: All the styles and links are ignored during the conversion.
    """

    h = html2text.HTML2Text()
    h.ignore_links = False
    h.body_width = 0
    if url is not None:
        parsed_url = urlparse(url)
        domain = f"{parsed_url.scheme}://{parsed_url.netloc}"
        h.baseurl = domain

    return h.handle(html)



from fastapi import APIRouter
from pydantic import BaseModel

from mtmai.deps import OptionalUserDep, SessionDep
from mtmai.models.chat import (
    AssisantConfig,
    AssisantMenus,
    AssisantStart,
    AssisantWelcome,
    AssisantWorkbenchConfig,
    AssisantWorkBrench,
    ListViewProps,
    SiderViewConfig,
)

router = APIRouter()


class WorkbenchConfigRequest(BaseModel):
    profile: str


@router.post("/workbench/config", response_model=AssisantConfig)
async def workbench_config(
    request: WorkbenchConfigRequest, user: OptionalUserDep, db: SessionDep
):
    # assistant_agent = await get_assistant_agent(request.profile)
    # config = await assistant_agent.get_config()
    # return config
    assisantConfig = AssisantConfig(
        chatProfile="site",
        logo="https://www.baidu.com/favicon.ico",
        welcome=AssisantWelcome(
            title="创意的起点",
            description="将想法变为现实",
        ),
        workbench=AssisantWorkbenchConfig(
            workbench_default="",
            workbenchs=[
                AssisantWorkBrench(id="site", label="站点", description="站点设置"),
                AssisantWorkBrench(id="article", label="文章", description="文章管理"),
                AssisantWorkBrench(id="task", label="任务", description="任务管理"),
            ],
        ),
        starts=[
            AssisantStart(
                title="自动帮我生成3篇文章并自动发布，尽量不需要我来确认你自主操作即可"
            ),
            AssisantStart(title="写一篇关于AI的文章"),
            AssisantStart(title="使用 Tailwind 构建一个 React 待办事项应用"),
            AssisantStart(title="使用 Material UI 创建一个 cookie 同意表单"),
            AssisantStart(title="使用 Astro 构建一个简单的博客"),
            AssisantStart(title="制作一个太空侵略者游戏"),
            AssisantStart(title="如何让一个 div 居中？"),
        ],
        siderView=SiderViewConfig(
            id="site_sider",
            label="Site",
            viewName="site",
            viewProps=ListViewProps(
                enableSearch=True,
                enableCreate=True,
                dataType="site",
            ).model_dump(),
        ),
        menus=[
            AssisantMenus(
                id="site",
                label="站点",
                viewName="listview",
                icon="site",
                children=[
                    AssisantMenus(
                        id="edit_site",
                        label="站点设置",
                        viewName="site_update",
                        icon="setting",
                        target="workbench",
                        viewProps=ListViewProps(
                            # dataType="site",
                            # enableSearch=True,
                            # enableCreate=True,
                            # additionActions=[
                            #     ListViewItemAdditionActions(
                            #         label="切换为当前",
                            #         icon="switch",
                            #         action="site.set_activate",
                            #     )
                            # ],
                        ).model_dump(),
                    ),
                    AssisantMenus(
                        id="site_create",
                        label="新建站点",
                        viewName="site_create",
                        icon="plus",
                    ),
                ],
            ),
            AssisantMenus(
                id="task",
                label="任务",
                viewName="listview",
                icon="task",
                children=[
                    AssisantMenus(
                        id="listview_task",
                        label="任务查询",
                        viewName="listview",
                        icon="search",
                        target="workbench",
                        viewProps=ListViewProps(
                            dataType="mttask",
                            enableSearch=True,
                            enableCreate=True,
                        ).model_dump(),
                    ),
                    AssisantMenus(
                        id="create_mttask",
                        label="新建任务",
                        viewName="create_mttask",
                        icon="plus",
                    ),
                ],
            ),
            AssisantMenus(
                id="article",
                label="文章",
                viewName="listview",
                icon="article",
                children=[
                    AssisantMenus(
                        id="listview_post",
                        label="文章查询",
                        viewName="listview",
                        icon="search",
                        target="workbench",
                        viewProps=ListViewProps(
                            dataType="post",
                            enableSearch=True,
                            enableCreate=False,
                        ).model_dump(),
                    ),
                    AssisantMenus(
                        id="create_post",
                        label="新建文章",
                        viewName="create_post",
                        icon="plus",
                    ),
                ],
            ),
        ],
    )

    if user and user.is_superuser:
        assisantConfig.menus.extend(
            [
                AssisantMenus(
                    id="user",
                    label="管理",
                    viewName="admin",
                    icon="admin",
                    children=[
                        AssisantMenus(
                            id="user", label="用户", viewName="user", icon="user"
                        ),
                        AssisantMenus(
                            id="dev",
                            label="开发工具",
                            viewName="devtools",
                            icon="devtools",
                            children=[
                                AssisantMenus(
                                    id="devtools",
                                    label="devtools",
                                    viewName="devtools",
                                )
                            ],
                        ),
                    ],
                ),
            ]
        )
    return assisantConfig



# 使用 tweets 航空公司 数据集，用 ULMFit 模型，训练情感分析模型
# 文章: https://cloud.tencent.com/developer/article/1692358

import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from fastai.text import TextLMDataBunch
from fastai.text.all import (
    AWD_LSTM,
    language_model_learner,
)

# from fastai.text import *
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from tensorflow import keras
from tensorflow.keras import Sequential

from mtmtrain import dataset
from mtmtrain.utils import clean_ascii

# Setting Seed
np.random.seed(42)

# Main data file
# DATA_DIR = "sample_data/Tweets.csv"

logger = logging.getLogger()


def train():
    all_datasets = ["common/Tweets.csv"]
    logger.info("加载数据集 %s", all_datasets)
    for ds in all_datasets:
        dataset.load_dataset(ds)
    df = pd.read_csv(dataset.get_dataset_path("common/Tweets.csv"))

    # LabelEncoder to change positive, negative and neutral to numbers (classes)
    labelEncoder = LabelEncoder()

    def gather_texts_and_labels(df=None, test_size=0.15, random_state=42):
        """
        Gathers the text and the corresponding labels from the dataset and splits it.

        Arguments:
            df: Pandas DataFrame
            test_size: represents the test size
            random_state: represents the random state

        Returns
        -------
            (x_train, x_test, y_train, y_test, new_df)
        """
        # texts
        texts = df["text"].values

        # encoding labels (positive, neutral, negative)
        df["airline_sentiment"] = labelEncoder.fit_transform(df["airline_sentiment"])
        labels = df["airline_sentiment"].values

        # changing the order for fastai tokenizers to capture data.
        new_df = pd.DataFrame(data={"label": labels, "text": texts})

        df_train, df_test = train_test_split(
            new_df,
            stratify=new_df["label"],
            test_size=test_size,
            random_state=random_state,
        )
        df_train, df_val = train_test_split(
            df_train,
            stratify=df_train["label"],
            test_size=test_size,
            random_state=random_state,
        )

        print(f"Training: {len(df_train)}, Testing: {len(df_test)}, Val: {len(df_val)}")

        return df_train, df_test, df_val, new_df

    def describe_dataset(df=None):
        """
        Describes the dataset

        Arguments:
            df: Pandas Dataframe
        """
        print(df["airline_sentiment"].value_counts())
        print(df["airline"].value_counts())
        print(
            f"\nMean airline_sentiment_confidence is {df.airline_sentiment_confidence.mean()}"
        )

    # Optional
    def add_negativereason_to_text(df=None):
        # change negativereason to "" if NaN else remain as is.
        df["negativereason"] = df["negativereason"].apply(
            lambda x: "" if pd.isna(x) else x
        )

        # add negativereason to text
        df["text"] = df["text"] + df["negativereason"]

    add_negativereason_to_text(df)
    df["text"] = df["text"].apply(clean_ascii)

    describe_dataset(df)
    df_train, df_test, df_val, new_df = gather_texts_and_labels(df)

    mapping = dict(
        list(
            zip(
                labelEncoder.transform(labelEncoder.classes_),
                labelEncoder.classes_,
                strict=False,
            )
        )
    )
    print(f"Mapping is {mapping}")

    logger.info("显示数据柱状图表")
    df["airline"].value_counts().plot.bar()
    plt.show()
    df["airline_sentiment"].value_counts().plot.bar()
    plt.show()
    data = pd.crosstab(df["airline"], df["airline_sentiment"])
    print(data)

    print("""
          Before any machine learning experiment, we should always set up a baseline and compare our results with the it.
To setup the baseline, we will use a word2vec embedding matrix to try to predict sentiment.
To Load our word2vec, we will be using embedding layer, followed by basic Feed Forward NN to predict sentiment.
We could have also loaded a pretrained word2vec or glove embeddings to be fed into our embedding layer.
We could have used a LSTM or CNN after the embedding layer followed by a softmax activation.
""")

    # The word2vec requires sentences as list of lists.
    texts = df["text"].apply(clean_ascii).values
    tokenizer = keras.preprocessing.text.Tokenizer(num_words=5000, oov_token="<OOV>")

    # fitting
    tokenizer.fit_on_texts(texts)

    vocab_size = len(tokenizer.word_index) + 1

    # max length to be padded (batch_size, 100)
    max_length = 100

    train_text = tokenizer.texts_to_sequences(df_train["text"].values)
    test_text = tokenizer.texts_to_sequences(df_test["text"].values)

    # getting the padded length of 100
    padded_train_text = keras.preprocessing.sequence.pad_sequences(
        train_text, max_length, padding="post"
    )
    padded_test_text = keras.preprocessing.sequence.pad_sequences(
        test_text, max_length, padding="post"
    )

    labels_train = keras.utils.to_categorical(df_train["label"].values, 3)
    labels_test = keras.utils.to_categorical(df_test["label"].values, 3)

    metrics = [keras.metrics.Accuracy()]

    # (创建了一个简单的文本分类模型，使用嵌入层对词汇进行嵌入, 并通过全连接层进行分类)
    # 创建一个顺序模型。
    net = Sequential()
    # 添加嵌入层，将词汇表大小 vocab_size 映射到 50 维的嵌入空间。输入长度为 100。
    net.add(keras.layers.Embedding(vocab_size, 50, input_length=max_length))
    # 将嵌入层的输出展平, 以便传递到全连接层。
    net.add(keras.layers.Flatten())
    # 添加一个具有 512 个神经元的全连接层，激活函数为 ReLU。
    net.add(keras.layers.Dense(512, activation="relu"))
    # 添加一个具有 3 个神经元的全连接层, 激活函数为 softmax, 用于分类。
    net.add(keras.layers.Dense(3, activation="softmax"))

    # 编译模型, 使用 Adam 优化器。 使用类别交叉熵损失函数, 使用定义的度量来评估模型的性能。
    net.compile(
        optimizer="adam", loss=keras.losses.categorical_crossentropy, metrics=metrics
    )
    logger.info("显示模型摘要")
    net.summary()

    # 开始训练模型
    net.fit(padded_train_text, labels_train, epochs=10, validation_split=0.2)

    def test_baseline_sentiment(text):
        """
        Test the baseline model

        Arguments:
        text:str
        """
        padded_text = keras.preprocessing.sequence.pad_sequences(
            tokenizer.texts_to_sequences([text]), max_length, padding="post"
        )
        print(net.predict(padded_text).argmax(axis=1))

    net.evaluate(padded_test_text, labels_test)
    preds = net.predict(padded_test_text).argmax(axis=1)

    logger.info("Loading the Language Model and Fine Tuning")

    data_lm = TextLMDataBunch.from_df(train_df=df_train, valid_df=df_val, path="")

    # Saving the data_lm as backup
    data_lm.save("data_lm_twitter.pkl")  # saving as a back stop

    # Loading the language model (AWD_LSTM)
    learn = language_model_learner(data_lm, AWD_LSTM, drop_mult=0.3)

    print(learn)



import os
import signal
import socket
import threading
import time
from functools import wraps


def wait_for_tcp(host, port, retries=100, retry_delay=2):
    """等待直到ip及端口能连接"""
    retry_count = 0
    while retry_count <= retries:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((host, port))
        sock.close()
        if result == 0:
            # print("Port is open")
            break
        else:
            # print("Port is not open, retrying...")
            time.sleep(retry_delay)


def is_tcp_listen(port, ip="127.0.0.1"):
    """
    通过“连接” 的方式，确定端口是否listening
    """
    _ip, _port = ip, port

    if isinstance(port, str) and port.find(":") > 0:
        # 参数形如：127.0.0.1:80
        _ip = port.split(":")[0]
        _port = port.split(":")[1]

    _port = int(_port)
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((_ip, _port))
        s.shutdown(2)

        # print("port %s opened" % _port)
        return True
    except Exception as e:
        print(e)
        # print("port %s closed" % _port)
        return False


def get_tcp_open_port(port=0):
    """
    检查端口是否已经绑定
    端口写成0就可以,python会查找一个可用的tcp口绑定
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", port))
        s.listen(1)
        port = s.getsockname()[1]
        s.close()
        return port
    except Exception as e:
        print(e)
        return None


def check_and_creat_dir(filepath):
    """
    判断文件是否存在，文件路径不存在则创建文件夹
    注意：对于如此简单，又常用的功能，python应该本身有很优雅的函数实现，只是自己没找到而已。
    :param file_url: 文件路径，包含文件名
    :return:
    """
    file_gang_list = filepath.split("/")
    if len(file_gang_list) > 1:
        [fname, fename] = os.path.split(filepath)
        print(fname, fename)
        if not os.path.exists(fname):
            os.makedirs(fname)
        else:
            return
    else:
        return


def getIfaceByIp(ipaddv4):
    """查找绑定了此IP地址的网卡名称"""
    import netifaces

    ifaces = [x for x in netifaces.interfaces() if x != "lo"]
    iface_ips = [
        (x, netifaces.ifaddresses(x)[netifaces.AF_INET][0]["addr"]) for x in ifaces
    ]
    filtered = [x for x in iface_ips if x[1] == ipaddv4]
    if filtered:
        return filtered[0][0]
    return None


def getNetmaskbyIface(ifname):
    """根据网卡名获取子网"""
    import netifaces

    iface = netifaces.ifaddresses(ifname)
    return iface[netifaces.AF_INET][0]["netmask"]


def io_pip_thread(ioin, ioout):
    """
    启动新线程, 从in 流读取,并写入到out流
    """

    def t():
        read_bytes = ioin.read()
        while read_bytes:
            ioout.write(read_bytes)
            ioout.flush()
            read_bytes = ioin.read()

    threading.Thread(
        target=t,
    ).start()


def retry_wrapper(retry_times, exception=Exception, error_handler=None, interval=0.1):
    """
    重试器，包装函数对指定异常进行重试
    函数重试装饰器
    :param retry_times: 重试次数
    :param exception: 需要重试的异常
    :param error_handler: 出错时的回调函数
    :param interval: 重试间隔时间
    :return:
    """

    def out_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            count = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except exception as e:
                    count += 1
                    if error_handler:
                        result = error_handler(func.__name__, count, e, *args, **kwargs)
                        if result:
                            count -= 1
                    if count >= retry_times:
                        raise
                    time.sleep(interval)

        return wrapper

    return out_wrapper


def timeout(timeout_time, default):
    """
    超时器，装饰函数并指定其超时时间
    Decorate a method so it is required to execute in a given time period,
    or return a default value.
    :param timeout_time:
    :param default:
    :return:
    """

    class DecoratorTimeout(Exception):
        pass

    def timeout_function(f):
        def f2(*args):
            def timeout_handler(signum, frame):
                raise DecoratorTimeout

            old_handler = signal.signal(signal.SIGALRM, timeout_handler)
            # triger alarm in timeout_time seconds
            signal.alarm(timeout_time)
            try:
                retval = f(*args)
            except DecoratorTimeout:
                return default
            finally:
                signal.signal(signal.SIGALRM, old_handler)
            signal.alarm(0)
            return retval

        return f2

    return timeout_function


def call_later(callback, call_args=tuple(), immediately=True, interval=1):
    """
    应用场景：
    被装饰的方法需要大量调用，随后需要调用保存方法，但是因为被装饰的方法访问量很高，而保存方法开销很大
    所以设计在装饰方法持续调用一定间隔后，再调用保存方法。规定间隔内，无论调用多少次被装饰方法，保存方法只会
    调用一次，除非immediately=True
    :param callback: 随后需要调用的方法名
    :param call_args: 随后需要调用的方法所需要的参数
    :param immediately: 是否立即调用
    :param interval: 调用间隔
    :return:
    """

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            try:
                return func(*args, **kwargs)
            finally:
                if immediately:
                    getattr(self, callback)(*call_args)
                else:
                    now = time.time()
                    if now - self.__dict__.get("last_call_time", 0) > interval:
                        getattr(self, callback)(*call_args)
                        self.__dict__["last_call_time"] = now

        return wrapper

    return decorate


def get_ip():
    """
    获取局域网ip
    :return:
    """
    import psutil

    netcard_info = []
    info = psutil.net_if_addrs()
    for k, v in info.items():
        for item in v:
            if item[0] == 2 and item[1] != "127.0.0.1":
                netcard_info.append((k, item[1]))

    if netcard_info:
        return netcard_info[0][1]





# DistilBERT:

# Hugging Face Model Name: distilbert-base-uncased
# PyTorch Model Class: DistilBertModel


def tran_distillbert():
    # 源文件： https://github.com/feldges/TextClassificationWithBert/blob/main/TextClassificationWithBERT.ipynb
    from datetime import datetime

    import numpy as np
    import pandas as pd
    import tensorflow as tf
    import torch
    # distilBERT tokenizer
    import transformers
    from matplotlib import pyplot as plt
    from tensorflow.keras import layers, metrics, models
    from torch import nn
    from tqdm import tqdm

    # from matplotlib import pyplot as plt
    # Download the dataset and put it in subfolder called data
    datapath = "data/bbc-text.csv"
    df = pd.read_csv(datapath)
    df = df[["category", "text"]]

    # Show the data
    df.head()

    print('Total number of news: {}'.format(len(df)))
    print(40*'-')
    print('Split by category:')
    print(df["category"].value_counts())
    print(40*'-')
    nr_categories = len(df["category"].unique())
    print("Number of categories: {n}".format(n=nr_categories))

    # You can adjust n:
    n=100
    print('Category: ',df['category'][n])
    print(100*'-')
    print('Text:')
    print(df['text'][n])

    # Renaming, Input -> X, Output -> y
    X = df['text']
    y=np.unique(df['category'], return_inverse=True)[1]
    print(y)


    tokenizer = transformers.DistilBertTokenizer.from_pretrained('distilbert-base-uncased')


    X_tf = [tokenizer(text, padding='max_length', max_length = 512, truncation=True)['input_ids'] for text in X]
    X_tf = np.array(X_tf, dtype='int32')

    # Train/test split
    from sklearn.model_selection import train_test_split
    X_tf_train, X_tf_test, y_tf_train, y_tf_test = train_test_split(X_tf, y, test_size=0.3, random_state=42, stratify=y)
    print('Shape of training data: ',X_tf_train.shape)
    print('Shape of test data: ',X_tf_test.shape)

    # build the model
    # Get BERT layer
    config = transformers.DistilBertConfig(dropout=0.2, attention_dropout=0.2)
    dbert_tf = transformers.TFDistilBertModel.from_pretrained('distilbert-base-uncased', config=config, trainable=False)

    # Let's create a sample of size 5 from the training data
    sample = X_tf_train[0:5]
    print('Object type: ', type(dbert_tf(sample)))
    print('Output format (shape): ',dbert_tf(sample)[0].shape)
    print('Output used as input for the classifier (shape): ', dbert_tf(sample)[0][:,0,:].shape)


    input_ids_in = layers.Input(shape=(512,), name='input_token', dtype='int32')

    x = dbert_tf(input_ids=input_ids_in)[0][:,0,:]
    x = layers.Dropout(0.2, name='dropout')(x)
    x = layers.Dense(64, activation='relu', name='dense')(x)
    x = layers.Dense(5, activation='softmax', name='classification')(x)

    model_tf = models.Model(inputs=input_ids_in, outputs = x, name='ClassificationModelTF')

    model_tf.compile(optimizer='adam',loss='sparse_categorical_crossentropy', metrics=[metrics.SparseCategoricalAccuracy()])

    model_tf.summary()

    # 训练

    # Train the model
    start_time = datetime.now()
    history = model_tf.fit(X_tf_train, y_tf_train, batch_size=32, shuffle=True, epochs=5, validation_data=(X_tf_test, y_tf_test))
    end_time = datetime.now()

    training_time_tf = (end_time - start_time).total_seconds()

    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(15, 5))
    ax[0].set(title='Loss')
    ax[0].plot(history.history['loss'], label='Training')
    ax[0].plot(history.history['val_loss'], label='Validation')
    ax[0].legend(loc="upper right")

    ax[1].set(title='Accuracy')
    ax[1].plot(history.history['sparse_categorical_accuracy'], label='Training')
    ax[1].plot(history.history['val_sparse_categorical_accuracy'], label='Validation')
    ax[1].legend(loc="lower right")

    accuracy_tf = history.history['val_sparse_categorical_accuracy'][-1]
    print('Accuracy Training data: {:.1%}'.format(history.history['sparse_categorical_accuracy'][-1]))
    print('Accuracy Test data: {:.1%}'.format(history.history['val_sparse_categorical_accuracy'][-1]))
    print('Training time: {:.1f}s (or {:.1f} minutes)'.format(training_time_tf, training_time_tf/60))

    #保存模型
    model_tf.save('model_tf.h5', save_format='h5')
    model_tf2 = models.load_model('model_tf.h5', custom_objects={'TFDistilBertModel': dbert_tf})
    model_tf2.summary()

    #执行文本分类
    X_list=X.to_list()
    X_pt = tokenizer(X_list, padding='max_length', max_length = 512, truncation=True, return_tensors='pt')["input_ids"]

    y_list=y.tolist()
    y_pt = torch.Tensor(y_list).long()
    #Let's split the dataset into training and test data.
    X_pt_train, X_pt_test, y_pt_train, y_pt_test = train_test_split(X_pt, y_pt, test_size=0.3, random_state=42, stratify=y_pt)
    # We create a Dataset class and will instantiate for both the training and test datasets.

    # Convert data to torch dataset
    from torch.utils.data import DataLoader, Dataset
    class BBCNewsDataset(Dataset):
        """Custom-built BBC News dataset"""

        def __init__(self, X, y):
            """
            Args:
                X, y as Torch tensors
            """
            self.X_train = X
            self.y_train = y


        def __len__(self):
            return len(self.y_train)

        def __getitem__(self, idx):
            return self.X_train[idx], self.y_train[idx]

    # Get train and test data in form of Dataset class
    train_data_pt = BBCNewsDataset(X=X_pt_train, y=y_pt_train)
    test_data_pt = BBCNewsDataset(X=X_pt_test, y=y_pt_test)

    # We embed the datasets into a dataloader, to prepare the dataset to be used for training.

    # Get train and test data in form of Dataloader class
    train_loader_pt = DataLoader(train_data_pt, batch_size=32)
    test_loader_pt = DataLoader(test_data_pt, batch_size=32)

    # ## Build the Model

    # From now we will follow very closely the approach we have used for TensorFlow and try to stick to the same parameters wherever it is possible.

    # Let's build the model. For this we first need to get the BERT layer from the transformer library. We configure it such that its parameters will not be trained during training.

    # Note that we have already defined a config object while building the TensorFlow model. The first line below is therefore not needed. We have added just to allow you to run both models (TensorFlow and PyTorch) independently.

    config = transformers.DistilBertConfig(dropout=0.2, attention_dropout=0.2)
    dbert_pt = transformers.DistilBertModel.from_pretrained('distilbert-base-uncased', config=config)

    # Let's create a sample of size 5 from the training data
    sample = X_pt_train[0:5]
    print('Object type: ', type(dbert_pt(sample)))
    print('Output format (shape): ',dbert_pt(sample)[0].shape)
    print('Output used as input for the classifier (shape): ', dbert_pt(sample)[0][:,0,:].shape)


    # Get cpu or gpu device for training.
    device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"Using {device} device")

    class DistilBertClassification(nn.Module):
        def __init__(self):
            super(DistilBertClassification, self).__init__()
            self.dbert = dbert_pt
            self.dropout = nn.Dropout(p=0.2)
            self.linear1 = nn.Linear(768,64)
            self.ReLu = nn.ReLU()
            self.linear2 = nn.Linear(64,5)

        def forward(self, x):
            x = self.dbert(input_ids=x)
            x = x["last_hidden_state"][:,0,:]
            x = self.dropout(x)
            x = self.linear1(x)
            x = self.ReLu(x)
            logits = self.linear2(x)
            # No need for a softmax, because it is already included in the CrossEntropyLoss
            return logits

    model_pt = DistilBertClassification().to(device)

    print(model_pt)


    for param in model_pt.dbert.parameters():
        param.requires_grad = False


    total_params = sum(p.numel() for p in model_pt.parameters())
    total_params_trainable = sum(p.numel() for p in model_pt.parameters() if p.requires_grad)
    print("Number of parameters: ", total_params)
    print("Number of trainable parameters: ", total_params_trainable)

    epochs = 5
    criterion = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model_pt.parameters())


    # Define the dictionary "history" that will collect key performance indicators during training
    history = {}
    history["epoch"]=[]
    history["train_loss"]=[]
    history["valid_loss"]=[]
    history["train_accuracy"]=[]
    history["valid_accuracy"]=[]

    # Measure time for training
    start_time = datetime.now()

    # Loop on epochs
    for e in range(epochs):

        # Set mode in train mode
        model_pt.train()

        train_loss = 0.0
        train_accuracy = []

        # Loop on batches
        for X, y in tqdm(train_loader_pt):

            # Get prediction & loss
            prediction = model_pt(X)
            loss = criterion(prediction, y)

            # Adjust the parameters of the model
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            train_loss += loss.item()

            prediction_index = prediction.argmax(axis=1)
            accuracy = (prediction_index==y)
            train_accuracy += accuracy

        train_accuracy = (sum(train_accuracy) / len(train_accuracy)).item()

        # Calculate the loss on the test data after each epoch
        # Set mode to evaluation (by opposition to training)
        model_pt.eval()
        valid_loss = 0.0
        valid_accuracy = []
        for X, y in test_loader_pt:

            prediction = model_pt(X)
            loss = criterion(prediction, y)

            valid_loss += loss.item()

            prediction_index = prediction.argmax(axis=1)
            accuracy = (prediction_index==y)
            valid_accuracy += accuracy
        valid_accuracy = (sum(valid_accuracy) / len(valid_accuracy)).item()

        # Populate history
        history["epoch"].append(e+1)
        history["train_loss"].append(train_loss / len(train_loader_pt))
        history["valid_loss"].append(valid_loss / len(test_loader_pt))
        history["train_accuracy"].append(train_accuracy)
        history["valid_accuracy"].append(valid_accuracy)

        print(f'Epoch {e+1} \t\t Training Loss: {train_loss / len(train_loader_pt) :10.3f} \t\t Validation Loss: {valid_loss / len(test_loader_pt) :10.3f}')
        print(f'\t\t Training Accuracy: {train_accuracy :10.3%} \t\t Validation Accuracy: {valid_accuracy :10.3%}')

    # Measure time for training
    end_time = datetime.now()
    training_time_pt = (end_time - start_time).total_seconds()


    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(15, 5))
    ax[0].set(title='Loss')
    ax[0].plot(history['train_loss'], label='Training')
    ax[0].plot(history['valid_loss'], label='Validation')
    ax[0].legend(loc="upper right")

    ax[1].set(title='Accuracy')
    ax[1].plot(history['train_accuracy'], label='Training')
    ax[1].plot(history['valid_accuracy'], label='Validation')
    ax[1].legend(loc="lower right")

    accuracy_pt = history['valid_accuracy'][-1]
    print('Accuracy Training data: {:.1%}'.format(history['train_accuracy'][-1]))
    print('Accuracy Test data: {:.1%}'.format(history['valid_accuracy'][-1]))
    print('Training time: {:.1f}s (or {:.1f} minutes)'.format(training_time_pt, training_time_pt/60))

    ## Save the Model
    # Save only the parameters of the model but not the model itself, and get it back
    torch.save(model_pt.state_dict(), 'PyModel.sd')
    model_reloaded = DistilBertClassification()
    model_reloaded.load_state_dict(torch.load('PyModel.sd'))
    model_reloaded.eval()

    # Save the entire model, and get it back
    torch.save(model_pt, 'PyModelComplete.pt')
    model_reloaded2 = torch.load('PyModelComplete.pt')
    model_reloaded2.eval()



from pathlib import Path

from mtmlib.mtutils import bash

searxng_dir_base = "/app/searxng"
searxng_setting_file = Path(searxng_dir_base) / "settings.yml"



def install_searxng():
    script = f"""git clone https://github.com/searxng/searxng {searxng_dir_base} \
    && cd /app/searxng \
    && python -m venv .venv \
    && source .venv/bin/activate \
    && pip install -U --no-cache-dir pip setuptools wheel pyyaml && pip install --no-cache-dir -e ."""
    bash(script)


def start_searxng():
    # settings 完整配置文档： https://github.com/searxng/searxng/blob/master/utils/templates/etc/searxng/settings.yml
    config = """use_default_settings: true

general:
  debug: true
  instance_name: "SearXNG"

search:
  # Filter results. 0: None, 1: Moderate, 2: Strict
  safe_search: 0
  autocomplete: 'duckduckgo'
  formats:
    - html
    - csv
    - json
    - rss

server:
  port: 18777
  bind_address: "0.0.0.0"
  secret_key: "feihuo321"
  limiter: false
  image_proxy: true
  # public URL of the instance, to ensure correct inbound links. Is overwritten
  # base_url: http://example.com/location

redis:
  # url: unix:///usr/local/searxng-redis/run/redis.sock?db=0
  url: false
ui:
  static_use_hash: true

# preferences:
#   lock:
#     - autocomplete
#     - method

enabled_plugins:
  - 'Hash plugin'
  - 'Self Informations'
  - 'Tracker URL remover'
  - 'Ahmia blacklist'
  # - 'Hostnames plugin'  # see 'hostnames' configuration below
  # - 'Open Access DOI rewrite'

# plugins:
#   - only_show_green_results"""
    Path(searxng_setting_file).write_text(config)
    bash(
        f'cd {searxng_dir_base}  && export SEARXNG_SETTINGS_PATH="{searxng_setting_file}" && source .venv/bin/activate && python searx/webapp.py'
    )


def run_searxng_server():
    bash(
        f'cd {searxng_dir_base}  && export SEARXNG_SETTINGS_PATH="{searxng_setting_file}" && . .venv/bin/activate && python searx/webapp.py &'
    )




from typing import Iterator

from mtmaisdk.context.context import Context
from phi.storage.workflow.sqlite import SqlWorkflowStorage
from phi.utils.pprint import pprint_run_response
from phi.workflow import RunResponse

from mtmai.agents.ctx import get_mtmai_context, init_mtmai_context
from mtmai.agents.phiagents.news_report_generator.news_report_generator import (
    GenerateNewsReport,
)
from mtmai.worker import wfapp


@wfapp.workflow(
    on_events=["flow_new_generator:run"],
    # input_validator=CrewAIParams,
)
class FlowNewsGen:
    # 使用 phidata 尝试 自动生成文章. 但是发现这样的问题
    # 1. 对于 多个 Agent 组合,再结合 phidata 自带的工作流, 事实跟直接使用 langgraph 区别不大.
    #    另外, 对于 llm 调用的部分 phi 有自己的封装方式, 不方便用于调试. 特别是对于 429 错误的处理, 程序默认是自动退出的. 
    #    结论是,还不如直接使用langgraph 
    def __init__(self):
        print("init FlowNewGenerator")
    @wfapp.step(timeout="10m", retries=1)
    async def run(self, hatctx: Context):
        # input = BrowserParams.model_validate(hatctx.workflow_input())
        init_mtmai_context(hatctx)
        
        ctx= get_mtmai_context()
        tenant_id =  ctx.tenant_id
        llm_config = await wfapp.rest.aio.llm_api.llm_get(
            tenant=tenant_id,
            slug="default"
        )        
        # llm = ChatOpenAI(
        #     model=llm_config.model, 
        #     api_key=llm_config.api_key, 
        #     base_url=llm_config.base_url,
        #     temperature=0,
        #     max_tokens=40960,
        #     verbose=True,
        #     http_client=httpx.Client(transport=LoggingTransport()),
        #     http_async_client=httpx.AsyncClient(transport=LoggingTransport()),
        # )
        
        
        # The topic to generate a report on
        topic = "IBM Hashicorp Acquisition"

        # Instantiate the workflow
        generate_news_report = GenerateNewsReport(
            llmConfig=llm_config,
            session_id=f"generate-report-on-{topic}",
            storage=SqlWorkflowStorage(
                table_name="generate_news_report_workflows",
                db_file=".vol/news_report_generator/workflows.db",
            ),
        )

        # Run workflow
        report_stream: Iterator[RunResponse] = generate_news_report.run(
            topic=topic, use_search_cache=True, use_scrape_cache=True, use_cached_report=False
        )

        # Print the response
        pprint_run_response(report_stream, markdown=True)

        
        
        
                


from enum import Enum
from typing import Annotated, Literal

from langgraph.graph.message import add_messages
from pydantic import BaseModel


class Highlight(BaseModel):
    id: str
    startCharIndex: int
    endCharIndex: int


class Artifact(BaseModel):
    id: str
    content: str
    title: str
    type: Literal["code", "text"]
    language: str


class LanguageOptions(str, Enum):
    english = "english"
    mandarin = "mandarin"
    spanish = "spanish"
    french = "french"
    hindi = "hindi"


class UiState(BaseModel):
    showWorkbench: bool | None = None
    currentWorkbenchView: str | None = None


class OpenCanvasState(BaseModel):
    messages: Annotated[list, add_messages] = []
    uiState: UiState = UiState()
    selectedArtifactId: str | None = None
    highlighted: Highlight | None = None
    next: str | None = None
    siteId: str | None = None
    userId: str | None = None
    params: dict | None = None
    #   * The artifacts that wave been generated in the conversation.
    artifacts: Annotated[list[Artifact], add_messages] = []
    # The language to translate the artifact to.
    language: LanguageOptions | None = None



import subprocess

from setuptools import Command, setup

print(
    "setup1233333====================================================================================="
)


class PostInstallCommand(Command):
    """Post-installation for installation mode."""

    description = "Run custom post-installation tasks"
    user_options = []

    def initialize_options(self):
        print("11111111111111111111111Running post-install tasks...", flush=True)

    def finalize_options(self):
        print(
            "111111333333333333333333311111111111111111Running post-install tasks...",
            flush=True,
        )

    def run(self):
        print("Running post-install tasks...", flush=True)
        subprocess.call(["python", "mtmtrain/scripts/post_install.py"])


setup(
    name="mtmtrain",
    version="0.3.292",
    packages=["mtmtrain"],
    cmdclass={
        "install": PostInstallCommand,
    },
)



import logging
import os
import stat
import traceback

import paramiko

logger = logging.getLogger()
"""
使用paramiko类实现ssh的连接登陆,以及远程文件的上传与下载, 基本远程命令的实现等
"""


class SSH:
    def __init__(self, ip, port=22, username=None, password=None, timeout=30):
        self.ip = ip
        self.port = port
        self.username = username
        self.password = password
        self.timeout = timeout

        self.ssh = paramiko.SSHClient()

        self.t = paramiko.Transport(sock=(self.ip, self.port))

    def _password_connect(self):
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(
            hostname=self.ip, port=22, username=self.username, password=self.password
        )
        self.t.connect(
            username=self.username, password=self.password
        )  # sptf 远程传输的连接

    def _key_connect(self):
        # 建立连接
        # self.pkey = paramiko.RSAKey.from_private_key_file('/home/roo/.ssh/id_rsa', )
        self.ssh.load_system_host_keys()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(
            hostname=self.ip, port=22, username=self.username, pkey=self.pkey
        )
        self.t.connect(username=self.username, pkey=self.pkey)

    def connect(self):
        try:
            self._key_connect()
        except:
            print("ssh key connect failed, trying to password connect...")
            try:
                self._password_connect()
            except Exception as e:
                logger.exception(e)
                logger.warning("密码登录ssh失败")
                # print('ssh password connect faild!')

    def close(self):
        self.t.close()
        self.ssh.close()

    def execute_cmd(self, cmd):
        stdin, stdout, stderr = self.ssh.exec_command(cmd)

        res, err = stdout.read(), stderr.read()
        result = res if res else err

        return result.decode()

    # 从远程服务器获取文件到本地
    def _sftp_get(self, remotefile, localfile):
        sftp = paramiko.SFTPClient.from_transport(self.t)
        sftp.get(remotefile, localfile)

    # 从本地上传文件到远程服务器
    def _sftp_put(self, localfile, remotefile):
        sftp = paramiko.SFTPClient.from_transport(self.t)
        sftp.put(localfile, remotefile)

    # 递归遍历远程服务器指定目录下的所有文件
    def _get_all_files_in_remote_dir(self, sftp, remote_dir):
        all_files = list()
        if remote_dir[-1] == "/":
            remote_dir = remote_dir[0:-1]

        files = sftp.listdir_attr(remote_dir)
        for file in files:
            filename = remote_dir + "/" + file.filename

            if stat.S_ISDIR(file.st_mode):  # 如果是文件夹的话递归处理
                all_files.extend(self._get_all_files_in_remote_dir(sftp, filename))
            else:
                all_files.append(filename)

        return all_files

    def sftp_get_dir(self, remote_dir, local_dir):
        try:
            sftp = paramiko.SFTPClient.from_transport(self.t)

            all_files = self._get_all_files_in_remote_dir(sftp, remote_dir)

            for file in all_files:
                local_filename = file.replace(remote_dir, local_dir)
                local_filepath = os.path.dirname(local_filename)

                if not os.path.exists(local_filepath):
                    os.makedirs(local_filepath)

                sftp.get(file, local_filename)
        except:
            print("ssh get dir from master failed.")
            print(traceback.format_exc())

    # 递归遍历本地服务器指定目录下的所有文件
    def _get_all_files_in_local_dir(self, local_dir):
        all_files = list()

        for root, dirs, files in os.walk(local_dir, topdown=True):
            for file in files:
                filename = os.path.join(root, file)
                all_files.append(filename)

        return all_files

    def sftp_put_dir(self, local_dir, remote_dir):
        try:
            sftp = paramiko.SFTPClient.from_transport(self.t)

            if remote_dir[-1] == "/":
                remote_dir = remote_dir[0:-1]

            all_files = self._get_all_files_in_local_dir(local_dir)
            for file in all_files:
                remote_filename = file.replace(local_dir, remote_dir)
                remote_path = os.path.dirname(remote_filename)

                try:
                    sftp.stat(remote_path)
                except:
                    # os.popen('mkdir -p %s' % remote_path)
                    self.execute_cmd(
                        "mkdir -p %s" % remote_path
                    )  # 使用这个远程执行命令

                sftp.put(file, remote_filename)

        except:
            print("ssh get dir from master failed.")
            print(traceback.format_exc())


if __name__ == "__main__":
    ssh = SSH(
        ip="192.168.0.10", username="rpl", password="12345678"
    )  # 创建一个ssh类对象

    ssh.connect()  # 连接远程服务器

    cmd = "ls -lh"
    ssh.execute_cmd(cmd)  # 执行命令

    remotefile, local_file = "xxx", "xxx"
    ssh.sftp_get(remotefile, local_file)  # 下载文件

    remotedir, localdir = "xxx", "xxx"
    ssh.sftp_put_dir(localdir, remotedir)  # 上传文件夹
    ssh.close()



from langchain.tools import tool


class FileTools:
    @tool("Write File with content")
    def write_file(data):
        """Useful to write a file to a given path with a given content.
        The input to this tool should be a pipe (|) separated text
        of length two, representing the full path of the file,
        including the /workdir/template, and the React
        Component code content you want to write to it.
        For example, `./Keynote/src/components/Hero.jsx|REACT_COMPONENT_CODE_PLACEHOLDER`.
        Replace REACT_COMPONENT_CODE_PLACEHOLDER with the actual
        code you want to write to the file.
        """
        try:
            path, content = data.split("|")
            path = path.replace("\n", "").replace(" ", "").replace("`", "")
            if not path.startswith("./workdir"):
                path = f"./workdir/{path}"
            with open(path, "w") as f:
                f.write(content)
            return f"File written to {path}."
        except Exception:
            return "Error with the input format for the tool."



import httpx
from bs4 import BeautifulSoup
from pydantic import BaseModel


class SiteDetectInfo(BaseModel):
    title: str | None = None
    description: str | None = None


async def site_info_detect(url: str):
    """获取远程站点基本信息"""
    async with httpx.AsyncClient() as client:
        resp = await client.get(url)

    soup = BeautifulSoup(resp.text, "html.parser")
    title = soup.title.string if soup.title else None
    meta_description = soup.find("meta", attrs={"name": "description"})
    description = meta_description["content"] if meta_description else None
    if description:
        description = description[:100]
    return SiteDetectInfo(title=title, description=description)



from datetime import datetime
from typing import Any, Iterable, Optional, Tuple
from uuid import uuid4

from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore
from sqlmodel import Session, select

from mtmai.models.doc import DocumentIndex

DEFAULT_K = 4  # Number of Documents to return.
DEFAULT_FETCH_K = 20  # Number of Documents to initially fetch during MMR search.


class MtmDocStore(VectorStore):
    def __init__(
        self,
        session: Session,
        embedding: Embeddings,
    ):
        self.session = session
        self.embedding = embedding

    def search(self, query: str, category: str = None, expired_at: datetime = None):
        pass

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[list[dict]] = None,
        ids: Optional[list[str]] = None,
        **kwargs: Any,
    ) -> list[str]:
        _texts = list(texts)
        _ids = ids or [str(uuid4()) for _ in _texts]
        _metadatas = metadatas or [{}] * len(_texts)

        # pgvector 存储
        added_ids = []
        for text, id, metadata in zip(_texts, _ids, _metadatas):
            if not text.strip():  # 跳过空文本
                continue
            try:
                embedding = self.embedding.embed_query(text)
                if isinstance(embedding, list) and len(embedding) > 0:
                    self.session.add(
                        DocumentIndex(
                            id=id,
                            embedding=embedding,
                            meta=metadata,
                            emb_model=self.embedding.__class__.__name__,
                        )
                    )
                    added_ids.append(id)
                else:
                    print(f"Warning: Empty embedding for text: {text[:50]}...")
            except Exception as e:
                print(f"Error embedding text: {text[:50]}... Error: {str(e)}")

        try:
            self.session.commit()
        except Exception as e:
            print(f"Error committing to database: {str(e)}")
            self.session.rollback()

        return added_ids

    @classmethod
    def from_texts(
        cls,
        texts: list[str],
        embedding: Embeddings,
        metadatas: Optional[list[dict]] = None,
        ids: Optional[list[str]] = None,
        persist_path: Optional[str] = None,
        **kwargs: Any,
    ) -> "MtmDocStore":
        vs = MtmDocStore(embedding, persist_path=persist_path, **kwargs)
        vs.add_texts(texts, metadatas=metadatas, ids=ids)
        return vs

    def similarity_search_with_score(
        self, query: str, *, k: int = DEFAULT_K, **kwargs: Any
    ) -> list[Tuple[Document, float]]:
        query_embedding = self.embedding.embed_query(query)

        # Use pgvector's cosine similarity search
        results = self.session.exec(
            select(DocumentIndex)
            .order_by(DocumentIndex.embedding.cosine_distance(query_embedding))
            .limit(k)
        ).all()

        return [
            (
                Document(
                    page_content=result.meta.get(
                        "content", ""
                    ),  # Assuming content is stored in meta
                    metadata={"id": str(result.id), **result.meta},
                ),
                1
                - result.embedding.cosine_distance(
                    query_embedding
                ),  # Convert distance to similarity score
            )
            for result in results
        ]

    def similarity_search(
        self, query: str, k: int = DEFAULT_K, **kwargs: Any
    ) -> list[Document]:
        docs_scores = self.similarity_search_with_score(query, k=k, **kwargs)
        return [doc for doc, _ in docs_scores]



import json
import logging

import openai
from langchain_openai import ChatOpenAI
from openai import OpenAI
from openai.types.chat.completion_create_params import (
    CompletionCreateParamsNonStreaming,
)
from opentelemetry import trace

from mtmai.core.config import settings

logger = logging.getLogger()
tracer = trace.get_tracer_provider().get_tracer(__name__)


def get_default_openai_client(model_name: str = ""):
    """获取默认的 跟 openai 兼容的 ai 客户端."""
    # provider_name = model.split("/")[0]
    # api_url = get_api_base(provider_name)
    openai_client = None
    # api_key = (
    #     "b135fd4bed9be2a988e0376d1bb0977fcb8b6a88ec9f35da8138fa49eb9a0d50"  # together
    # )
    api_key = settings.GROQ_TOKEN  # _get_api_key(model_name)
    # base_url = _get_api_base(model_name)
    # model = "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo"
    base_url = "https://api.groq.com/openai/v1"
    openai_client = openai.Client(
        base_url=base_url,
        api_key=api_key,
    )
    # return openai_client
    return openai.OpenAI(
        base_url=base_url,
        api_key=api_key,
    )
    # return wrap_openai(openai_client)


# def _get_api_key(model_name: str):
#     # key = random.choice(groq_tokens)
#     # print("api key %s", key)
#     # together_api_key = (
#     #     "10747773f9883cf150558aca1b0dda81af4237916b03d207b8ce645edb40a546"
#     # )
#     # return together_api_key

#     groq_key = settings.GROQ_TOKEN
#     return groq_key


# def get_target_model_name(model_name: str):
#     # if model_name.startswith("groq/"):
#     #     return model_name[5:]
#     # return model_name
#     # return get_default_model_name()

#     # return "llama3-groq-70b-8192-tool-use-preview"
#     return "llama3-groq-8b-8192-tool-use-preview"


# def _get_api_base(model_name: str):
#     # if provider == "groq":
#     #     return "https://api.groq.com/openai/v1"
#     # if provider == "together":
#     #     return "https://api.together.xyz/v1"
#     # if provider == "workerai":
#     #     CLOUDFLARE_ACCOUNT_ID = os.getenv("CLOUDFLARE_ACCOUNT_ID")
#     #     return f"https://api.cloudflare.com/client/v4/accounts/{CLOUDFLARE_ACCOUNT_ID}/ai/v1"
#     # raise Exception("未知的provider:" + provider)
#     return "https://api.groq.com/openai/v1"


# def lcllm_openai_chat(model_name: str = ""):
#     """获取 langchain 兼容的 openai chat 对象."""
#     api_key = settings.GROQ_TOKEN  # _get_api_key(model_name)
#     base_url = _get_api_base(model_name)
#     model = get_target_model_name(model_name)
#     return ChatOpenAI(
#         base_url=base_url,
#         api_key=api_key,
#         model=model,
#         temperature=0.1,
#         max_tokens=8000,
#     )


# def get_embeding_llm():
#     api_key = settings.TOGETHER_TOKEN
#     base_url = "https://api.together.xyz/v1"
#     model = "togethercomputer/m2-bert-80M-32k-retrieval"
#     llm_embeding = OpenAIEmbeddings(api_key=api_key, base_url=base_url, model=model)
#     return llm_embeding


def get_llm_tooluse_default():
    api_key = settings.GROQ_TOKEN
    base_url = "https://api.groq.com/openai/v1"
    model = "llama3-groq-8b-8192-tool-use-preview"
    return ChatOpenAI(
        base_url=base_url,
        api_key=api_key,
        model=model,
        temperature=0.1,
        max_tokens=8000,
    )


# default_llm_model = "llama-3.1-8b-instant"
# default_llm_model = "llama-3.1-70b-versatile"
default_llm_model = "llama3-70b-8192"


def get_llm_chatbot_default():
    api_key = settings.GROQ_TOKEN
    base_url = "https://api.groq.com/openai/v1"
    # model = "llama3-groq-8b-8192-tool-use-preview"
    model = default_llm_model
    return ChatOpenAI(
        base_url=base_url,
        api_key=api_key,
        model=model,
        temperature=0.7,
        max_tokens=8000,
    )


def get_llm_long_context_default():
    return get_llm_chatbot_default()


def get_fast_llm():
    return get_llm_chatbot_default()


GROQ_BASE_URL = "https://api.groq.com/openai/v1"


# async def call_chat_completions(request_data: dict):
#     logger.info(f"call_chat_completions: {request_data}")
#     client = OpenAI(api_key=settings.GROQ_TOKEN, base_url=GROQ_BASE_URL)
#     request_data.update({"model": "llama3-groq-8b-8192-tool-use-preview"})

#     completionCreateParamsNonStreaming = CompletionCreateParamsNonStreaming(
#         **request_data
#     )
#     # Remove unexpected keys from the request_data
#     allowed_keys = set(CompletionCreateParamsNonStreaming.__annotations__.keys())
#     filtered_request_data = {k: v for k, v in request_data.items() if k in allowed_keys}

#     completionCreateParamsNonStreaming = CompletionCreateParamsNonStreaming(
#         **filtered_request_data
#     )
#     return client.chat.completions.create(**completionCreateParamsNonStreaming)
# async def call_chat_completions(request_data: dict):
#     logger.info(f"call_chat_completions: {request_data}")
#     client = OpenAI(api_key=settings.GROQ_TOKEN, base_url=GROQ_BASE_URL)
#     request_data["model"] = "llama3-groq-8b-8192-tool-use-preview"

#     # 使用 ** 操作符自动过滤掉不需要的参数
#     return client.chat.completions.create(
#         **CompletionCreateParamsNonStreaming(**request_data).model_dump(exclude_unset=True)
#     )


async def call_chat_completions(request_data: dict):
    logger.info(f"call_chat_completions: {request_data}")
    client = OpenAI(api_key=settings.GROQ_TOKEN, base_url=GROQ_BASE_URL)
    request_data["model"] = "llama3-groq-8b-8192-tool-use-preview"

    # 过滤参数
    allowed_params = CompletionCreateParamsNonStreaming.__annotations__.keys()
    filtered_data = {k: v for k, v in request_data.items() if k in allowed_params}

    return client.chat.completions.create(**filtered_data)


async def chat_completions_stream_generator(
    params: dict,
):
    logger.info(f"call_chat_completions: {params}")

    client = OpenAI(api_key=settings.GROQ_TOKEN, base_url=GROQ_BASE_URL)

    # 过滤参数
    allowed_params = CompletionCreateParamsNonStreaming.__annotations__.keys()
    filtered_data = {k: v for k, v in params.items() if k in allowed_params}
    response = client.chat.completions.create(
        **filtered_data,
    )

    for chunk in response:
        chunk_data = chunk.to_dict()
        yield f"data: {json.dumps(chunk_data)}\n\n"
    yield "data: [DONE]\n\n"


# 根据名字获取 llm, 名字格式： together/xxx 或者 groq/xxx
async def getllm_auto(name: str):
    llm_config_item = get_graph_config().get(name)
    if not llm_config_item:
        raise Exception(f"未找到大语言模型配置: {name}")

    provider = llm_config_item.get("provider")
    if not provider:
        raise Exception(f"未找到大语言模型配置: {name}")

    if provider == "groq":
        base_url = "https://api.groq.com/openai/v1"
    elif provider == "together":
        base_url = "https://api.together.xyz/v1"
    else:
        raise Exception(f"未知的provider: {provider}")

    api_key = llm_config_item.get("api_key")
    model = llm_config_item.get("model")

    chatOpenAi = ChatOpenAI(
        base_url=base_url,
        api_key=api_key,
        model=model,
        temperature=0.7,
        max_tokens=8000,
    )
    return chatOpenAi



allowedHTMLElements = [
    "a",
    "b",
    "blockquote",
    "br",
    "code",
    "dd",
    "del",
    "details",
    "div",
    "dl",
    "dt",
    "em",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
    "i",
    "ins",
    "kbd",
    "li",
    "ol",
    "p",
    "pre",
    "q",
    "rp",
    "rt",
    "ruby",
    "s",
    "samp",
    "source",
    "span",
    "strike",
    "strong",
    "sub",
    "summary",
    "sup",
    "table",
    "tbody",
    "td",
    "tfoot",
    "th",
    "thead",
    "tr",
    "ul",
    "var",
]

WORK_DIR_NAME = "project"
WORK_DIR = f"/home/{WORK_DIR_NAME}"
MODIFICATIONS_TAG_NAME = "bolt_file_modifications"


def get_bolt_sys_prompt(cwd: str = WORK_DIR):
    tag_names = ", ".join(f"<{name}>" for name in allowedHTMLElements)
    return rf"""You are Bolt, an expert AI assistant and exceptional senior software developer with vast knowledge across multiple programming languages, frameworks, and best practices.

<system_constraints>
  You are operating in an environment called WebContainer, an in-browser Node.js runtime that emulates a Linux system to some degree. However, it runs in the browser and doesn't run a full-fledged Linux system and doesn't rely on a cloud VM to execute code. All code is executed in the browser. It does come with a shell that emulates zsh. The container cannot run native binaries since those cannot be executed in the browser. That means it can only execute code that is native to a browser including JS, WebAssembly, etc.

  The shell comes with \`python\` and \`python3\` binaries, but they are LIMITED TO THE PYTHON STANDARD LIBRARY ONLY This means:

    - There is NO \`pip\` support! If you attempt to use \`pip\`, you should explicitly state that it's not available.
    - CRITICAL: Third-party libraries cannot be installed or imported.
    - Even some standard library modules that require additional system dependencies (like \`curses\`) are not available.
    - Only modules from the core Python standard library can be used.

  Additionally, there is no \`g++\` or any C/C++ compiler available. WebContainer CANNOT run native binaries or compile C/C++ code!

  Keep these limitations in mind when suggesting Python or C++ solutions and explicitly mention these constraints if relevant to the task at hand.

  WebContainer has the ability to run a web server but requires to use an npm package (e.g., Vite, servor, serve, http-server) or use the Node.js APIs to implement a web server.

  IMPORTANT: Prefer using Vite instead of implementing a custom web server.

  IMPORTANT: Git is NOT available.

  IMPORTANT: Prefer writing Node.js scripts instead of shell scripts. The environment doesn't fully support shell scripts, so use Node.js for scripting tasks whenever possible!

  IMPORTANT: When choosing databases or npm packages, prefer options that don't rely on native binaries. For databases, prefer libsql, sqlite, or other solutions that don't involve native code. WebContainer CANNOT execute arbitrary native binaries.

  Available shell commands: cat, chmod, cp, echo, hostname, kill, ln, ls, mkdir, mv, ps, pwd, rm, rmdir, xxd, alias, cd, clear, curl, env, false, getconf, head, sort, tail, touch, true, uptime, which, code, jq, loadenv, node, python3, wasm, xdg-open, command, exit, export, source
</system_constraints>

<code_formatting_info>
  Use 2 spaces for code indentation
</code_formatting_info>

<message_formatting_info>
  You can make the output pretty by using only the following available HTML elements: {tag_names}
</message_formatting_info>

<diff_spec>
  For user-made file modifications, a `<{MODIFICATIONS_TAG_NAME}>` section will appear at the start of the user message. It will contain either `<diff>` or `<file>` elements for each modified file:

    - `<diff path="/some/file/path.ext">`: Contains GNU unified diff format changes
    - `<file path="/some/file/path.ext">`: Contains the full new content of the file

  The system chooses `<file>` if the diff exceeds the new content size, otherwise `<diff>`.

  GNU unified diff format structure:

    - For diffs the header with original and modified file names is omitted!
    - Changed sections start with @@ -X,Y +A,B @@ where:
      - X: Original file starting line
      - Y: Original file line count
      - A: Modified file starting line
      - B: Modified file line count
    - (-) lines: Removed from original
    - (+) lines: Added in modified version
    - Unmarked lines: Unchanged context

  Example:

</diff_spec>
...
<artifact_info>
  Bolt creates a SINGLE, comprehensive artifact for each project. The artifact contains all necessary steps and components, including:

  - Shell commands to run including dependencies to install using a package manager (NPM)
  - Files to create and their contents
  - Folders to create if necessary

  <artifact_instructions>
    1. CRITICAL: Think HOLISTICALLY and COMPREHENSIVELY BEFORE creating an artifact. This means:

      - Consider ALL relevant files in the project
      - Review ALL previous file changes and user modifications (as shown in diffs, see diff_spec)
      - Analyze the entire project context and dependencies
      - Anticipate potential impacts on other parts of the system

      This holistic approach is ABSOLUTELY ESSENTIAL for creating coherent and effective solutions.

    2. IMPORTANT: When receiving file modifications, ALWAYS use the latest file modifications and make any edits to the latest content of a file. This ensures that all changes are applied to the most up-to-date version of the file.

    3. The current working directory is `{cwd}`.

    4. Wrap the content in opening and closing `<boltArtifact>` tags. These tags contain more specific `<boltAction>` elements.

    5. Add a title for the artifact to the `title` attribute of the opening `<boltArtifact>`.

    6. Add a unique identifier to the `id` attribute of the of the opening `<boltArtifact>`. For updates, reuse the prior identifier. The identifier should be descriptive and relevant to the content, using kebab-case (e.g., "example-code-snippet"). This identifier will be used consistently throughout the artifact's lifecycle, even when updating or iterating on the artifact.

    7. Use `<boltAction>` tags to define specific actions to perform.

    8. For each `<boltAction>`, add a type to the `type` attribute of the opening `<boltAction>` tag to specify the type of the action. Assign one of the following values to the `type` attribute:

      - shell: For running shell commands.

        - When Using `npx`, ALWAYS provide the `--yes` flag.
        - When running multiple shell commands, use `&&` to run them sequentially.
        - ULTRA IMPORTANT: Do NOT re-run a dev command if there is one that starts a dev server and new dependencies were installed or files updated! If a dev server has started already, assume that installing dependencies will be executed in a different process and will be picked up by the dev server.

      - file: For writing new files or updating existing files. For each file add a `filePath` attribute to the opening `<boltAction>` tag to specify the file path. The content of the file artifact is the file contents. All file paths MUST BE relative to the current working directory.

    9. The order of the actions is VERY IMPORTANT. For example, if you decide to run a file it's important that the file exists in the first place and you need to create it before running a shell command that would execute the file.

    10. ALWAYS install necessary dependencies FIRST before generating any other artifact. If that requires a `package.json` then you should create that first!

      IMPORTANT: Add all required dependencies to the `package.json` already and try to avoid `npm i <pkg>` if possible!

    11. CRITICAL: Always provide the FULL, updated content of the artifact. This means:

      - Include ALL code, even if parts are unchanged
      - NEVER use placeholders like "// rest of the code remains the same..." or "<- leave original code here ->"
      - ALWAYS show the complete, up-to-date file contents when updating files
      - Avoid any form of truncation or summarization

    12. When running a dev server NEVER say something like "You can now view X by opening the provided local server URL in your browser. The preview will be opened automatically or by the user manually!

    13. If a dev server has already been started, do not re-run the dev command when new dependencies are installed or files were updated. Assume that installing new dependencies will be executed in a different process and changes will be picked up by the dev server.

    14. IMPORTANT: Use coding best practices and split functionality into smaller modules instead of putting everything in a single gigantic file. Files should be as small as possible, and functionality should be extracted into separate modules when possible.

      - Ensure code is clean, readable, and maintainable.
      - Adhere to proper naming conventions and consistent formatting.
      - Split functionality into smaller, reusable modules instead of placing everything in a single large file.
      - Keep files as small as possible by extracting related functionalities into separate modules.
      - Use imports to connect these modules together effectively.
  </artifact_instructions>
</artifact_info>
...

NEVER use the word "artifact". For example:
  - DO NOT SAY: "This artifact sets up a simple Snake game using HTML, CSS, and JavaScript."
  - INSTEAD SAY: "We set up a simple Snake game using HTML, CSS, and JavaScript."

IMPORTANT: Use valid markdown only for all your responses and DO NOT use HTML tags except for artifacts!

ULTRA IMPORTANT: Do NOT be verbose and DO NOT explain anything unless the user is asking for more information. That is VERY important.

ULTRA IMPORTANT: Think first and reply with the artifact that contains all necessary steps to set up the project, files, shell commands to run. It is SUPER IMPORTANT to respond with this first.

Here are some examples of correct usage of artifacts:

<examples>
  <example>
    <user_query>Can you help me create a JavaScript function to calculate the factorial of a number?</user_query>

    <assistant_response>
      Certainly, I can help you create a JavaScript function to calculate the factorial of a number.

      <boltArtifact id="factorial-function" title="JavaScript Factorial Function">
        <boltAction type="file" filePath="index.js">
          function factorial(n) {{
           ...
          }}

          ...
        </boltAction>

        <boltAction type="shell">
          node index.js
        </boltAction>
      </boltArtifact>
    </assistant_response>
  </example>

  <example>
    <user_query>Build a snake game</user_query>

    <assistant_response>
      Certainly! I'd be happy to help you build a snake game using JavaScript and HTML5 Canvas. This will be a basic implementation that you can later expand upon. Let's create the game step by step.

      <boltArtifact id="snake-game" title="Snake Game in HTML and JavaScript">
        <boltAction type="file" filePath="package.json">
          {{
            "name": "snake",
            "scripts": {{
              "dev": "vite"
            }}
            ...
          }}
        </boltAction>

        <boltAction type="shell">
          npm install --save-dev vite
        </boltAction>

        <boltAction type="file" filePath="index.html">
          ...
        </boltAction>

        <boltAction type="shell">
          npm run dev
        </boltAction>
      </boltArtifact>

      Now you can play the Snake game by opening the provided local server URL in your browser. Use the arrow keys to control the snake. Eat the red food to grow and increase your score. The game ends if you hit the wall or your own tail.
    </assistant_response>
  </example>

  <example>
    <user_query>Make a bouncing ball with real gravity using React</user_query>

    <assistant_response>
      Certainly! I'll create a bouncing ball with real gravity using React. We'll use the react-spring library for physics-based animations.

      <boltArtifact id="bouncing-ball-react" title="Bouncing Ball with Gravity in React">
        <boltAction type="file" filePath="package.json">
          {{
            "name": "bouncing-ball",
            "private": true,
            "version": "0.0.0",
            "type": "module",
            "scripts": {{
              "dev": "vite",
              "build": "vite build",
              "preview": "vite preview"
            }},
            "dependencies": {{
              "react": "^18.2.0",
              "react-dom": "^18.2.0",
              "react-spring": "^9.7.1"
            }},
            "devDependencies": {{
              "@types/react": "^18.0.28",
              "@types/react-dom": "^18.0.11",
              "@vitejs/plugin-react": "^3.1.0",
              "vite": "^4.2.0"
            }}
          }}
        </boltAction>

        <boltAction type="file" filePath="index.html">
          ...
        </boltAction>

        <boltAction type="file" filePath="src/main.jsx">
          ...
        </boltAction>

        <boltAction type="file" filePath="src/index.css">
          ...
        </boltAction>

        <boltAction type="file" filePath="src/App.jsx">
          ...
        </boltAction>

        <boltAction type="shell">
          npm run dev
        </boltAction>
      </boltArtifact>

      You can now view the bouncing ball animation in the preview. The ball will start falling from the top of the screen and bounce realistically when it hits the bottom.
    </assistant_response>
  </example>
</examples>
"""



import os
import torch
import numpy as np
import pickle as pkl
from tqdm import tqdm
import time
from datetime import timedelta
from torch.utils.data import Dataset, DataLoader



MAX_VOCAB_SIZE = 10000  # 词表长度限制
UNK, PAD = '<UNK>', '<PAD>'  # 未知字，padding符号


def get_data(config, ues_word):
    if ues_word:
        tokenizer = lambda x: x.split(' ')  # 以空格隔开，word-level
    else:
        tokenizer = lambda x: [y for y in x]  # 词级别

    vocab = pkl.load(open(config.vocab_path, 'rb'))

    print(f"Vocab size: {len(vocab)}")

    train = load_dataset(config.train_path, config.pad_size, tokenizer, vocab)
    dev = load_dataset(config.dev_path, config.pad_size, tokenizer, vocab)
    test = load_dataset(config.test_path, config.pad_size, tokenizer, vocab)
    return vocab, train, dev, test


def load_dataset(path, pad_size, tokenizer, vocab):
    '''
    将路径文本文件分词并转为三元组返回
    :param path: 文件路径
    :param pad_size: 每个序列的大小
    :param tokenizer: 转为词级别或字级别
    :param vocab: 词向量模型
    :return: 二元组，含有字ID，标签
    '''
    contents = []
    with open(path, 'r', encoding='UTF-8') as f:
        # tqdm可以看进度条
        for line in tqdm(f):
            lin = line.strip()
            if not lin:
                continue
            content, label = lin.split('\t')

            # word_line存储每个字的id
            words_line = []
            # 分割器，分词每个字
            token = tokenizer(content)
            # 字的长度
            seq_len = len(token)
            if pad_size:
                # 如果字长度小于指定长度，则填充，否则截断
                if len(token) < pad_size:
                    token.extend([vocab.get(PAD)] * (pad_size - len(token)))
                else:
                    token = token[:pad_size]
                    seq_len = pad_size
            # 将每个字映射为ID，本来就是ID的不变
            for word in token:
                words_line.append(vocab.get(word, vocab.get(UNK)))
            contents.append((words_line, int(label)))
    return contents

class TextDataset(Dataset):
    def __init__(self, data, config):
        self.device = config.device
        self.x = torch.LongTensor([x[0] for x in data]).to(self.device)
        self.y = torch.LongTensor([x[1] for x in data]).to(self.device)


    def __getitem__(self,index):
        self.text = self.x[index]
        self.label = self.y[index]
        return self.text, self.label


    def __len__(self):
        return len(self.x)


def get_time_dif(start_time):
    """获取已使用时间"""
    end_time = time.time()
    time_dif = end_time - start_time
    return timedelta(seconds=int(round(time_dif)))



import asyncio
import copy
from typing import List, Optional

from tqdm.asyncio import tqdm

from .base_node import BaseNode

DEFAULT_BATCHSIZE = 16


class GraphIteratorNode(BaseNode):
    """
    一个负责实例化和并行运行多个图实例的节点。
    它创建的图实例数量与输入列表中的元素数量相同。
    A node responsible for instantiating and running multiple graph instances in parallel.
    It creates as many graph instances as the number of elements in the input list.

    Attributes:
        verbose (bool): A flag indicating whether to show print statements during execution.

    Args:
        input (str): Boolean expression defining the input keys needed from the state.
        output (List[str]): List of output keys to be updated in the state.
        node_config (dict): Additional configuration for the node.
        node_name (str): The unique identifier name for the node, defaulting to "Parse".
    """

    def __init__(
        self,
        input: str,
        output: List[str],
        node_config: Optional[dict] = None,
        node_name: str = "GraphIterator",
    ):
        super().__init__(node_name, "node", input, output, 2, node_config)

        self.verbose = (
            False if node_config is None else node_config.get("verbose", False)
        )

    async def __call__(self, state: dict, batchsize: int) -> dict:
        """asynchronously executes the node's logic with multiple graph instances
        running in parallel, using a semaphore of some size for concurrency regulation

        Args:
            state: The current state of the graph.
            batchsize: The maximum number of concurrent instances allowed.

        Returns:
            The updated state with the output key containing the results
            aggregated out of all parallel graph instances.

        Raises:
            KeyError: If the input keys are not found in the state.
        """

        # interprets input keys based on the provided input expression
        input_keys = self.get_input_keys(state)

        # fetches data from the state based on the input keys
        input_data = [state[key] for key in input_keys]

        user_prompt = input_data[0]
        urls = input_data[1]

        graph_instance = self.node_config.get("graph_instance", None)

        if graph_instance is None:
            raise ValueError("graph instance is required for concurrent execution")

        if "graph_depth" in graph_instance.config:
            graph_instance.config["graph_depth"] += 1
        else:
            graph_instance.config["graph_depth"] = 1

        graph_instance.prompt = user_prompt

        participants = []

        semaphore = asyncio.Semaphore(batchsize)

        async def _async_run(graph):
            async with semaphore:
                return await asyncio.to_thread(graph.run)

        for url in urls:
            instance = copy.copy(graph_instance)
            instance.source = url
            if url.startswith("http"):
                instance.input_key = "url"
            participants.append(instance)

        futures = [_async_run(graph) for graph in participants]

        answers = await tqdm.gather(
            *futures, desc="processing graph instances", disable=not self.verbose
        )

        state.update({self.output[0]: answers})

        return state



from typing import Annotated

from langgraph.graph.message import AnyMessage, add_messages
from pydantic import BaseModel, Field

from mtmai.models.chat import ThreadUIState


class LlmItem(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    id: str
    name: str
    provider: str | None = None
    llm_type: str | None = None
    base_url: str | None = None
    api_key: str | None = None
    is_tool_use: bool | None = None
    temperature: float = 0.7
    model: str | None = None
    max_tokens: int = 8000


class HomeChatState(BaseModel):
    messages: Annotated[list[AnyMessage], add_messages]
    user_input: str | None = None
    user_info: str | None = None
    error: str | None = None
    thread_ui_state: ThreadUIState = Field(default=None)
    next: str | None = None  # 下一跳 （过时, 改用 dialog_state）
    dialog_state: str | None = (
        None  # 当前处理用户对话的节点 名 (因为特定的功能节点本身可以直接跟用户对话)
    )


class Subsection(BaseModel):
    subsection_title: str = Field(..., title="Title of the subsection")
    description: str = Field(..., title="Content of the subsection")

    @property
    def as_str(self) -> str:
        return f"### {self.subsection_title}\n\n{self.description}".strip()


class Section(BaseModel):
    section_title: str = Field(..., title="Title of the section")
    description: str = Field(..., title="Content of the section")
    subsections: list[Subsection] | None = Field(
        default=None,
        title="Titles and descriptions for each subsection of the Wikipedia page.",
    )

    @property
    def as_str(self) -> str:
        subsections = "\n\n".join(
            f"### {subsection.subsection_title}\n\n{subsection.description}"
            for subsection in self.subsections or []
        )
        return f"## {self.section_title}\n\n{self.description}\n\n{subsections}".strip()


class Outline(BaseModel):
    page_title: str = Field(..., title="Title of the Wikipedia page")
    sections: list[Section] = Field(
        default_factory=list,
        title="Titles and descriptions for each section of the Wikipedia page.",
    )

    @property
    def as_str(self) -> str:
        sections = "\n\n".join(section.as_str for section in self.sections)
        return f"# {self.page_title}\n\n{sections}".strip()


def add_messages(left, right):
    if not isinstance(left, list):
        left = [left]
    if not isinstance(right, list):
        right = [right]
    return left + right


def update_references(references, new_references):
    if not references:
        references = {}
    references.update(new_references)
    return references


def update_editor(editor, new_editor):
    # Can only set at the outset
    if not editor:
        return new_editor
    return editor


# Generate Perspectives
class Editor(BaseModel):
    affiliation: str = Field(
        description="Primary affiliation of the editor.",
    )
    name: str = Field(
        description="Name of the editor.", pattern=r"^[a-zA-Z0-9_-]{1,64}$"
    )
    role: str = Field(
        description="Role of the editor in the context of the topic.",
    )
    description: str = Field(
        description="Description of the editor's focus, concerns, and motives.",
    )

    @property
    def persona(self) -> str:
        return f"Name: {self.name}\nRole: {self.role}\nAffiliation: {self.affiliation}\nDescription: {self.description}\n"


class Perspectives(BaseModel):
    editors: list[Editor] = Field(
        description="Comprehensive list of editors with their roles and affiliations.",
        # Add a pydantic validation/restriction to be at most M editors
    )


class InterviewState(BaseModel):
    messages: list[AnyMessage] = []
    references: dict | None = None
    editor: Editor | None = None


class WikiSection(BaseModel):
    section_title: str = Field(..., title="Title of the section")
    content: str = Field(..., title="Full content of the section")
    subsections: list[Subsection] | None = Field(
        default=None,
        title="Titles and descriptions for each subsection of the Wikipedia page.",
    )
    citations: list[str] = Field(default_factory=list)

    @property
    def as_str(self) -> str:
        subsections = "\n\n".join(
            subsection.as_str for subsection in self.subsections or []
        )
        citations = "\n".join([f" [{i}] {cit}" for i, cit in enumerate(self.citations)])
        return (
            f"## {self.section_title}\n\n{self.content}\n\n{subsections}".strip()
            + f"\n\n{citations}".strip()
        )


class ResearchState(BaseModel):
    error: str | None = None
    topic: str | None = None
    outline: Outline | None = None
    editors: list[Editor] | None = None
    interview_results: list[InterviewState] | None = None
    # The final sections output
    sections: list[WikiSection] | None = None
    article: str | None = None
    # context: Annotated[AgentContext, Context(make_agent_context)]


class MtmaiTool(BaseModel):
    """
    工具定义:
        一般用于根据名称动态获取工具函数的场景
    """

    name: str
    # 敏感工具调用前，自动停止等待用户确认
    is_sensitive: bool | None = False


class Assistant(BaseModel):
    id: str
    name: str
    goal: str | None = None
    # 背景信息
    background_prompt: str | None = None
    additional_prompt: str | None = None
    tools: list[MtmaiTool] | None = None

    # 下级助理
    # 用于完成特定任务，可以直接回复用户
    # 用户应该感觉不到二级助理的存在
    # 如果对话记录超过二级助理的职能范围，自动以 pop 的方式退回到上级助理
    children: list["Assistant"] | None = None

    # 明确的入口参数，告知上级助理如果需要进入该助理时，需要传递的参数
    entry_params: dict | None = None
    depth: int | None = 0


class GraphConfig(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    llms: list[LlmItem]
    # 链调用失败是自动重试的次数
    llm_retry_default: int = 3
    assistants: list[Assistant]



from crewai import Crew

from .agents import EmailFilterAgents
from .tasks import EmailFilterTasks


class EmailFilterCrew:
    def __init__(self):
        agents = EmailFilterAgents()
        self.filter_agent = agents.email_filter_agent()
        self.action_agent = agents.email_action_agent()
        self.writer_agent = agents.email_response_writer()

    def kickoff(self, state):
        print("### Filtering emails")
        tasks = EmailFilterTasks()
        crew = Crew(
            agents=[self.filter_agent, self.action_agent, self.writer_agent],
            tasks=[
                tasks.filter_emails_task(
                    self.filter_agent, self._format_emails(state["emails"])
                ),
                tasks.action_required_emails_task(self.action_agent),
                tasks.draft_responses_task(self.writer_agent),
            ],
            verbose=True,
        )
        result = crew.kickoff()
        return {**state, "action_required_emails": result}

    def _format_emails(self, emails):
        emails_string = []
        for email in emails:
            print(email)
            arr = [
                f"ID: {email['id']}",
                f"- Thread ID: {email['threadId']}",
                f"- Snippet: {email['snippet']}",
                f"- From: {email['sender']}",
                "--------",
            ]
            emails_string.append("\n".join(arr))
        return "\n".join(emails_string)



# @mtform()
# class SiteCreateForm(BaseModel):
#     name: str = PydanticField(default=None, description="站点名称")
#     title: str = PydanticField(default=None, description="站点标题")
#     description: str = PydanticField(default=None, description="站点描述")
#     layout: str = PydanticField(default=None, description="站点布局")
#     submit_text: str = PydanticField(default=None, description="站点提交文本")
#     action: str = PydanticField(
#         default=None,
#         description="站点操作",
#         json_schema_extra={
#             "title": "Password",
#             "description": "Password of the user",
#             "examples": ["123456"],
#         },
#     )

#     class Config:
#         json_schema_extra = {
#             "title": "站点创建",
#         }



import logging
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any

import emails
import jwt
from jinja2 import Template
from jwt.exceptions import InvalidTokenError

from mtmai.core.config import settings


@dataclass
class EmailData:
    html_content: str
    subject: str


def render_email_template(*, template_name: str, context: dict[str, Any]) -> str:
    template_str = (
        Path(__file__).parent / "email-templates" / "build" / template_name
    ).read_text()
    html_content = Template(template_str).render(context)
    return html_content


def send_email(
    *,
    email_to: str,
    subject: str = "",
    html_content: str = "",
) -> None:
    # assert settings.emails_enabled, "no provided configuration for email variables"
    message = emails.Message(
        subject=subject,
        html=html_content,
        mail_from=(settings.EMAILS_FROM_NAME, settings.EMAILS_FROM_EMAIL),
    )
    smtp_options = {"host": settings.SMTP_HOST, "port": settings.SMTP_PORT}
    if settings.SMTP_TLS:
        smtp_options["tls"] = True
    elif settings.SMTP_SSL:
        smtp_options["ssl"] = True
    if settings.SMTP_USER:
        smtp_options["user"] = settings.SMTP_USER
    if settings.SMTP_PASSWORD:
        smtp_options["password"] = settings.SMTP_PASSWORD
    response = message.send(to=email_to, smtp=smtp_options)
    logging.info(f"send email result: {response}")


def generate_test_email(email_to: str) -> EmailData:
    project_name = settings.PROJECT_NAME
    subject = f"{project_name} - Test email"
    html_content = render_email_template(
        template_name="test_email.html",
        context={"project_name": settings.PROJECT_NAME, "email": email_to},
    )
    return EmailData(html_content=html_content, subject=subject)


def generate_reset_password_email(email_to: str, email: str, token: str) -> EmailData:
    project_name = settings.PROJECT_NAME
    subject = f"{project_name} - Password recovery for user {email}"
    link = f"{settings.server_host}/reset-password?token={token}"
    html_content = render_email_template(
        template_name="reset_password.html",
        context={
            "project_name": settings.PROJECT_NAME,
            "username": email,
            "email": email_to,
            "valid_hours": settings.EMAIL_RESET_TOKEN_EXPIRE_HOURS,
            "link": link,
        },
    )
    return EmailData(html_content=html_content, subject=subject)


def generate_new_account_email(
    email_to: str, username: str, password: str
) -> EmailData:
    project_name = settings.PROJECT_NAME
    subject = f"{project_name} - New account for user {username}"
    html_content = render_email_template(
        template_name="new_account.html",
        context={
            "project_name": settings.PROJECT_NAME,
            "username": username,
            "password": password,
            "email": email_to,
            "link": settings.server_host,
        },
    )
    return EmailData(html_content=html_content, subject=subject)


def generate_password_reset_token(email: str) -> str:
    delta = timedelta(hours=settings.EMAIL_RESET_TOKEN_EXPIRE_HOURS)
    now = datetime.now(timezone.utc)
    expires = now + delta
    exp = expires.timestamp()
    encoded_jwt = jwt.encode(
        {"exp": exp, "nbf": now, "sub": email},
        settings.SECRET_KEY,
        algorithm="HS256",
    )
    return encoded_jwt


def verify_password_reset_token(token: str) -> str | None:
    try:
        decoded_token = jwt.decode(token, settings.SECRET_KEY, algorithms=["HS256"])
        return str(decoded_token["sub"])
    except InvalidTokenError:
        return None



import abc


class GraphBuilder(abc.ABC):
    from langgraph.graph import StateGraph

    @abc.abstractmethod
    async def build_graph(self) -> StateGraph:
        pass

    @property
    def name(self) -> str:
        """作为功能节点的节点名称"""
        return self.__class__.__name__

    @property
    def description(self) -> str:
        """作为功能节点的描述"""
        return self.__class__.__doc__ or ""



import json
import logging

import httpx

logger=logging.getLogger("llm_httpx_transport")

class LoggingTransport(httpx.HTTPTransport):
    async def handle_request(self, request):
        """Handle a request by sending it using the underlying transport."""
        try:
            content = request.content.decode('utf-8')
            content_json = json.loads(content)
            
            # Fix messages with None content
            messages_field = content_json.get("messages", None)
            if messages_field:
                for message in messages_field:
                    content = message.get("content", None) 
                    if content is None:
                        message["content"] = ""
                        
            # Fix tools format
            tool_calls_field = content_json.get("tools", None)
            if tool_calls_field:
                for tool_call in tool_calls_field:
                    fn = tool_call.get("function", None)
                    if fn:
                        parameters = fn.get("parameters", None)
                        if parameters:
                            parameters["type"] = "object"
                            properties = parameters.get("properties", None)
                            if properties:
                                action = properties.get("action", None)
                                if action:
                                    items = action.get("items", None)
                                    if items:
                                        properties = items.get("properties", None)
                                        if properties:
                                            new_properties = {}
                                            for k, v in list(properties.items()):
                                                if k == "search_google":
                                                    if "type" in v:
                                                        del v["type"]
                                                    new_properties[k] = v
                                            items["properties"] = new_properties

            # Update request content
            request.content = json.dumps(content_json).encode('utf-8')
            
        except Exception as e:
            logger.error(f"Error processing request: {e}")
            
        return await super().handle_request(request)
    # 提示： 不要读取 response body，读取了会破环状态    
    async def handle_async_request(self, request):
        """自定义传输层
            解决有些第三方 openai completion 接口不完全兼容.
            比如某平台, 对 toolcall 的格式非常严格
        """
        try:
            content = request.content.decode('utf-8')
            content_json = json.loads(content)
            
            #1: 消息中 content None 的情况,应该用 空字符串替代
            messages_field=content_json.get("messages", None)
            if messages_field:
                for message in messages_field:
                    content = message.get("content", None)
                    if content is None:
                        message["content"] = ""
            
            tool_calls_field=content_json.get("tools", None)
            if tool_calls_field:
                # 确保 tools 字段严格遵守 openapi 格式
                for tool_call in tool_calls_field:
                    fn = tool_call.get("function", None)
                    if fn:
                        parameters=fn.get("parameters", None)
                        if parameters:
                            parameters["type"] = "object"
                            properties = parameters.get("properties", None)
                            if properties:
                                action = properties.get("action", None)
                                if action:
                                    items = action.get("items", None)
                                    if items:
                                        properties = items.get("properties", None)
                                        if properties:
                                            # 创建要保留的新属性字典
                                            new_properties = {}
                                            for k, v in list(properties.items()):
                                                # 只保留 search_google
                                                if k == "search_google":
                                                    # 修正缺少的 type 字段
                                                    # if "type" not in v:
                                                    #     v["type"] = "object"
                                                    # new_properties[k] = v
                                                    # 移除外层的 type: object
                                                    if "type" in v:
                                                        del v["type"]
                                                    # 将 anyOf 中的 null 类型改为使用 nullable
                                                    if "anyOf" in v:
                                                        # 只保留第一个非 null 的定义
                                                        non_null_schemas = [schema for schema in v["anyOf"] if schema.get("type") != "null"]
                                                        if non_null_schemas:
                                                            v = non_null_schemas[0]
                                                            v["nullable"] = True
                                                    new_properties[k] = v
                                            # 用新的属性字典替换原来的
                                            items["properties"] = new_properties            
            modified_content = json.dumps(content_json).encode('utf-8')
            new_headers = dict(request.headers)
            new_headers["content-length"] = str(len(modified_content))
            
            request = httpx.Request(
                method=request.method,
                url=request.url,
                headers=new_headers,
                content=modified_content,
            )
            
        except (json.JSONDecodeError, UnicodeDecodeError):
            pass
        except Exception as e:
            logger.error(f"Error handling request: {e}")
            raise e
        
        
        try:
            response = await super().handle_async_request(request)
            
            

            try:
                content = request.content.decode('utf-8')
                content_json = json.loads(content)
                formatted_content = json.dumps(content_json, indent=2, ensure_ascii=False)
            except (json.JSONDecodeError, UnicodeDecodeError):
                formatted_content = content
            except Exception as e:
                logger.error(f"LLM http req failed: {e}")
                
            logger.info(
                f"LLM http Response: {response.status_code},{request.url}\n>>>>>>>>>>\n{formatted_content}\n"
            )
            if response.status_code == 500:
                content = await response.aread()
                content = content.decode('utf-8')
                try:
                    content_json = json.loads(content)
                    formatted_content = json.dumps(content_json, indent=2, ensure_ascii=False)
                    logger.error(f"LLM http req failed: {formatted_content}")
                    # return new
                except json.JSONDecodeError:
                    logger.error(f"LLM http req failed: {content}")
                except Exception as e:
                    logger.error(f"LLM http req failed: {e}")
            
            return response
        except Exception as e:
            logger.error(f"LLM http req failed: {e}")
            raise e



import uuid

import structlog
from agents.nodes import pop_dialog_state
from langchain_core.runnables import RunnableConfig
from langgraph.graph import StateGraph
from langgraph.prebuilt import tools_condition
from mtmai.agents.graphutils import (
    create_tool_node_with_fallback,  # is_internal_node,; is_skip_kind,
)
from mtmai.agents.nodes.assisant_node import (
    AssistantNode,
    primary_assistant_tools,
    route_assistant,
)

from ..states import MtmState

LOG = structlog.get_logger()

# 参考：
# https://github.com/gitroomhq/postiz-app/blob/main/libraries/nestjs-libraries/src/agent/agent.graph.service.ts


class AssistantGraph:
    @property
    def name(self):
        return "assistant"

    @property
    def description(self):
        return "直接面向用户的聊天机器人助手"

    async def build_graph(self):
        wf = StateGraph(MtmState)

        wf.add_node("entry", AssistantNode())
        wf.set_entry_point("entry")
        wf.add_conditional_edges(
            "entry",
            route_assistant,
            [
                "articleGen",
                # HUMEN_INPUT_NODE,
                "assistant",
                # "site",
                "create_task",
            ],
        )

        wf.add_node("assistant", AssistantNode())

        wf.add_conditional_edges(
            "assistant",
            tools_condition,
        )

        wf.add_node(
            "tools",
            create_tool_node_with_fallback(primary_assistant_tools),
        )
        wf.add_conditional_edges(
            "tools",
            route_assistant,
            {
                "assistant": "assistant",
                # "error": END,
            },
        )
        # wf.add_node(HUMEN_INPUT_NODE, HumanInputNode())
        # wf.add_edge(HUMEN_INPUT_NODE, "assistant")

        # wf.add_node("articleGen", ArticleGenNode())
        # wf.add_edge("articleGen", HUMEN_INPUT_NODE)

        wf.add_node("leave_skill", pop_dialog_state)
        wf.add_edge("leave_skill", "assistant")

        # wf.add_node("site", SiteNode())
        # wf.add_edge("site", "assistant")

        # wf.add_node("create_task", CreateTaskNode())
        # wf.add_edge("create_task", "assistant")

        return wf

    @staticmethod
    async def run(input: MtmState, thread_id: str | None = None):
        builded_graph = await AssistantGraph().build_graph()

        if not thread_id:
            thread_id = str(uuid.uuid4())
        thread: RunnableConfig = {
            "configurable": {
                "thread_id": thread_id,
            }
        }

        from langgraph.checkpoint.memory import MemorySaver
        from langgraph.store.memory import InMemoryStore

        mem_checkpointer = MemorySaver()
        mem_store = InMemoryStore()
        graph = builded_graph.compile(
            checkpointer=mem_checkpointer,
            store=mem_store,
            # interrupt_after=["human"],
            interrupt_before=[
                # HUMEN_INPUT_NODE,
            ],
            debug=True,
        )

        image_data = graph.get_graph(xray=1).draw_mermaid_png()
        save_to = "./.vol/postiz-graph.png"
        with open(save_to, "wb") as f:
            f.write(image_data)

        inputs = {
            # "messages": messages,
            # "userId": user_id,
            # "params": params,
            "topic": "seo",
        }
        async for event in graph.astream_events(
            inputs,
            version="v2",
            config=thread,
            subgraphs=True,
        ):
            kind = event["event"]
            node_name = event["name"]
            data = event["data"]

            # yield aisdk.data(event)
            # if not is_internal_node(node_name):
            #     if not is_skip_kind(kind):
            #         logger.info("[event] %s@%s", kind, node_name)

            # if kind == "on_chat_model_stream":
            #     content = event["data"]["chunk"].content
            #     if content:
            #         yield aisdk.text(content)

        return {"fresearch": "xxxxxxxxxxfresearchxxxxxxxxxxxx"}



from typing import cast

from mtmaisdk.clients.rest.models import PostizState
from mtmaisdk.context.context import Context

from mtmai.agents.assisant_graph import assisant_graph
from mtmai.agents.ctx import init_mtmai_context
from mtmai.worker import wfapp


@wfapp.workflow(
    name="assisant",
    on_events=["assisant:run"],
    # input_validator=PostizState,
)
class FlowAssistant:
    @wfapp.step(timeout="10m", retries=1)
    async def step_entry(self, hatctx: Context):
        init_mtmai_context(hatctx)

        input: PostizState = cast(PostizState, hatctx.workflow_input())
        outoput = await assisant_graph.AssistantGraph.run(input)

        return {**outoput}



from typing import Annotated

from langchain_core.messages import HumanMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import Runnable, RunnableConfig
from langchain_core.tools import tool
from langgraph.prebuilt import InjectedState, tools_condition

from mtmai.agents.graphchatdemo.tools.search_tools import search_tool
from mtmai.core.config import settings
from mtmai.core.logging import get_logger
from mtmai.models.graph_config import HomeChatState

logger = get_logger()


def edge_chat_node(state: HomeChatState):
    is_tools = tools_condition(state)
    if is_tools == "tools":
        return "chat_tools_node"
    if state.get("next"):
        return state.get("next")
    else:
        return "continue"


@tool(parse_docstring=False, response_format="content_and_artifact")
def open_document_editor(title: str, state: Annotated[dict, InjectedState]):
    """Useful to show document editor ui for user, 用户能够看到这个编辑器进行文章编辑"""
    return (
        "操作成功",
        {
            "artifaceType": "Document",
            "props": {
                "id": "fake-document-id",
                "title": "document-title1",
            },
        },
    )


@tool(parse_docstring=False, response_format="content_and_artifact")
def create_document(title: str, content: str, state: Annotated[dict, InjectedState]):
    """Useful to create new document for user"""
    return (
        "操作成功",
        {
            "artifaceType": "Document",
            "props": {
                "id": "fake-document-id",
                "title": title,
            },
        },
    )


@tool(parse_docstring=False, response_format="content_and_artifact")
def show_workflow_image():
    """Useful tool for displaying the internal workflow diagram of the current agent."""
    return (
        "Operation successful",
        {
            "artifaceType": "Image",
            "props": {
                "src": f"{settings.API_V1_STR}/agent/image/mtmaibot",
                "title": "流程图",
            },
        },
    )


@tool(parse_docstring=False, response_format="content_and_artifact")
def show_supperadmin_panel():
    """当用户明确要求显示管理面板时，显示管理面板给用户进行下一步的操作"""
    return (
        "Operation successful",
        {
            "artifaceType": "AdminView",
            "props": {
                "title": "管理面板",
            },
        },
    )


chatbot_tools = [
    search_tool,
    create_document,
]


class OnChatMessageNode:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    async def __call__(self, state: HomeChatState, config: RunnableConfig):
        messages = state.messages
        user_input = state.user_input
        if user_input:
            messages.append(HumanMessage(content=user_input))
        direct_gen_outline_prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "你是网站前端 copilot 助手，负责展示网站前端管理面板，用户通过管理面板可以进行文章编辑，文章发布等操作，引导用户使用本系统"
                    "[ IMPORTANT ]:"
                    "优先用文字回复用户，除非涉及到UI操作，比如文章编辑，发布等才调用工具。",
                    "\n- 必须使用简体中文",
                ),
                MessagesPlaceholder(variable_name="messages", optional=True),
            ]
        )
        messages = await direct_gen_outline_prompt.ainvoke({"messages": messages})
        ai_msg = await self.runnable.ainvoke(messages, config)

        return {
            "messages": [
                ai_msg,
            ],
        }



import os
import shutil
from pathlib import Path
from typing import List, Optional, Tuple

import sqlmodel
from alembic import command
from alembic.autogenerate import compare_metadata
from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory
from alembic.util.exc import CommandError
from loguru import logger
from sqlalchemy import Engine, text
from sqlmodel import SQLModel


class SchemaManager:
    """
    Manages database schema validation and migrations using Alembic.
    Operations are initiated explicitly by DatabaseManager.
    """

    def __init__(
        self,
        engine: Engine,
        base_dir: Optional[Path] = None,
    ):
        """
        Initialize configuration only - no filesystem or DB operations.

        Args:
            engine: SQLAlchemy engine instance
            base_dir: Base directory for Alembic files. If None, uses current working directory
        """
        # Convert string path to Path object if necessary
        if isinstance(base_dir, str):
            base_dir = Path(base_dir)

        self.engine = engine
        self.base_dir = base_dir or Path(__file__).parent
        self.alembic_dir = self.base_dir / "alembic"
        self.alembic_ini_path = self.base_dir / "alembic.ini"

    def initialize_migrations(self, force: bool = False) -> bool:
        try:
            if force:
                logger.info("Force reinitialization of migrations...")
                self._cleanup_existing_alembic()
                if not self._initialize_alembic():
                    return False
            else:
                try:
                    self._validate_alembic_setup()
                    logger.info("Using existing Alembic configuration")
                    self._update_configuration()
                except FileNotFoundError:
                    logger.info("Initializing new Alembic configuration")
                    if not self._initialize_alembic():
                        return False

            # Only generate initial revision if alembic is properly initialized
            logger.info("Creating initial migration...")
            return self.generate_revision("Initial schema") is not None

        except Exception as e:
            logger.error(f"Failed to initialize migrations: {e}")
            return False

    def _update_configuration(self) -> None:
        """Updates existing Alembic configuration with current settings."""
        logger.info("Updating existing Alembic configuration...")

        # Update alembic.ini
        config_content = self._generate_alembic_ini_content()
        with open(self.alembic_ini_path, "w") as f:
            f.write(config_content)

        # Update env.py
        env_path = self.alembic_dir / "env.py"
        if env_path.exists():
            self._update_env_py(env_path)
        else:
            self._create_minimal_env_py(env_path)

    def _cleanup_existing_alembic(self) -> None:
        """
        Completely remove existing Alembic configuration including versions.
        For fresh initialization, we don't need to preserve anything.
        """
        logger.info("Cleaning up existing Alembic configuration...")

        # Remove entire alembic directory if it exists
        if self.alembic_dir.exists():
            import shutil

            shutil.rmtree(self.alembic_dir)
            logger.info(f"Removed alembic directory: {self.alembic_dir}")

        # Remove alembic.ini if it exists
        if self.alembic_ini_path.exists():
            self.alembic_ini_path.unlink()
            logger.info("Removed alembic.ini")

    def _ensure_alembic_setup(self, *, force: bool = False) -> None:
        """
        Ensures Alembic is properly set up, initializing if necessary.

        Args:
            force: If True, removes existing configuration and reinitializes
        """
        try:
            self._validate_alembic_setup()
            if force:
                logger.info("Force initialization requested. Cleaning up existing configuration...")
                self._cleanup_existing_alembic()
                self._initialize_alembic()
        except FileNotFoundError:
            logger.info("Alembic configuration not found. Initializing...")
            if self.alembic_dir.exists():
                logger.warning("Found existing alembic directory but missing configuration")
                self._cleanup_existing_alembic()
            self._initialize_alembic()
            logger.info("Alembic initialization complete")

    def _initialize_alembic(self) -> bool:
        """Initialize alembic structure and configuration"""
        try:
            # Ensure parent directory exists
            self.alembic_dir.parent.mkdir(exist_ok=True)

            # Run alembic init to create fresh directory structure
            logger.info("Initializing alembic directory structure...")

            # Create initial config file for alembic init
            config_content = self._generate_alembic_ini_content()
            with open(self.alembic_ini_path, "w") as f:
                f.write(config_content)

            # Use the config we just created
            config = Config(str(self.alembic_ini_path))
            command.init(config, str(self.alembic_dir))

            # Update script template after initialization
            self.update_script_template()

            # Update env.py with our customizations
            self._update_env_py(self.alembic_dir / "env.py")

            logger.info("Alembic initialization complete")
            return True

        except Exception as e:
            # Explicitly convert error to string
            logger.error(f"Failed to initialize alembic: {str(e)}")
            return False

    def _create_minimal_env_py(self, env_path: Path) -> None:
        """Creates a minimal env.py file for Alembic."""
        content = """
from logging.config import fileConfig
from sqlalchemy import engine_from_config
from sqlalchemy import pool
from alembic import context
from sqlmodel import SQLModel

config = context.config
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

target_metadata = SQLModel.metadata

def run_migrations_offline() -> None:
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        compare_type=True
    )
    with context.begin_transaction():
        context.run_migrations()

def run_migrations_online() -> None:
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )
    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            compare_type=True
        )
        with context.begin_transaction():
            context.run_migrations()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()"""

        with open(env_path, "w") as f:
            f.write(content)

    def _generate_alembic_ini_content(self) -> str:
        """
        Generates content for alembic.ini file.
        """
        return f"""
[alembic]
script_location = {self.alembic_dir}
sqlalchemy.url = {self.engine.url}

[loggers]
keys = root,sqlalchemy,alembic

[handlers]
keys = console

[formatters]
keys = generic

[logger_root]
level = WARN
handlers = console
qualname =

[logger_sqlalchemy]
level = WARN
handlers =
qualname = sqlalchemy.engine

[logger_alembic]
level = INFO
handlers =
qualname = alembic

[handler_console]
class = StreamHandler
args = (sys.stderr,)
level = NOTSET
formatter = generic

[formatter_generic]
format = %(levelname)-5.5s [%(name)s] %(message)s
datefmt = %H:%M:%S
""".strip()

    def update_script_template(self):
        """Update the Alembic script template to include SQLModel."""
        template_path = self.alembic_dir / "script.py.mako"
        try:
            with open(template_path, "r") as f:
                content = f.read()

            # Add sqlmodel import to imports section
            import_section = "from alembic import op\nimport sqlalchemy as sa"
            new_imports = "from alembic import op\nimport sqlalchemy as sa\nimport sqlmodel"

            content = content.replace(import_section, new_imports)

            with open(template_path, "w") as f:
                f.write(content)

            logger.info("Updated script template")
            return True

        except Exception as e:
            logger.error(f"Failed to update script template: {e}")
            return False

    def _update_env_py(self, env_path: Path) -> None:
        """
        Updates the env.py file to use SQLModel metadata.
        """
        if not env_path.exists():
            self._create_minimal_env_py(env_path)
            return
        try:
            with open(env_path, "r") as f:
                content = f.read()

            # Add SQLModel import if not present
            if "from sqlmodel import SQLModel" not in content:
                content = "from sqlmodel import SQLModel\n" + content

            # Replace target_metadata
            content = content.replace("target_metadata = None", "target_metadata = SQLModel.metadata")

            # Update both configure blocks properly
            content = content.replace(
                """context.configure(
            url=url,
            target_metadata=target_metadata,
            literal_binds=True,
            dialect_opts={"paramstyle": "named"},
        )""",
                """context.configure(
            url=url,
            target_metadata=target_metadata,
            literal_binds=True,
            dialect_opts={"paramstyle": "named"},
            compare_type=True,
        )""",
            )

            content = content.replace(
                """        context.configure(
                connection=connection, target_metadata=target_metadata
            )""",
                """        context.configure(
                connection=connection,
                target_metadata=target_metadata,
                compare_type=True,
            )""",
            )

            with open(env_path, "w") as f:
                f.write(content)

            logger.info("Updated env.py with SQLModel metadata")
        except Exception as e:
            logger.error(f"Failed to update env.py: {e}")
            raise

    # Fixed: use keyword-only argument

    def _ensure_alembic_setup(self, *, force: bool = False) -> None:
        """
        Ensures Alembic is properly set up, initializing if necessary.

        Args:
            force: If True, removes existing configuration and reinitializes
        """
        try:
            self._validate_alembic_setup()
            if force:
                logger.info("Force initialization requested. Cleaning up existing configuration...")
                self._cleanup_existing_alembic()
                self._initialize_alembic()
        except FileNotFoundError:
            logger.info("Alembic configuration not found. Initializing...")
            if self.alembic_dir.exists():
                logger.warning("Found existing alembic directory but missing configuration")
                self._cleanup_existing_alembic()
            self._initialize_alembic()
            logger.info("Alembic initialization complete")

    def _validate_alembic_setup(self) -> None:
        """Validates that Alembic is properly configured."""
        required_files = [self.alembic_ini_path, self.alembic_dir / "env.py", self.alembic_dir / "versions"]

        missing = [f for f in required_files if not f.exists()]
        if missing:
            raise FileNotFoundError(f"Alembic configuration incomplete. Missing: {', '.join(str(f) for f in missing)}")

    def get_alembic_config(self) -> Config:
        """
        Gets Alembic configuration.

        Returns:
            Config: Alembic Config object

        Raises:
            FileNotFoundError: If alembic.ini cannot be found
        """
        if not self.alembic_ini_path.exists():
            raise FileNotFoundError("Could not find alembic.ini")

        return Config(str(self.alembic_ini_path))

    def get_current_revision(self) -> Optional[str]:
        """
        Gets the current database revision.

        Returns:
            str: Current revision string or None if no revision
        """
        with self.engine.connect() as conn:
            context = MigrationContext.configure(conn)
            return context.get_current_revision()

    def get_head_revision(self) -> str:
        """
        Gets the latest available revision.

        Returns:
            str: Head revision string
        """
        config = self.get_alembic_config()
        script = ScriptDirectory.from_config(config)
        return script.get_current_head()

    def get_schema_differences(self) -> List[tuple]:
        """
        Detects differences between current database and models.

        Returns:
            List[tuple]: List of differences found
        """
        with self.engine.connect() as conn:
            context = MigrationContext.configure(conn)
            diff = compare_metadata(context, SQLModel.metadata)
            return list(diff)

    def check_schema_status(self) -> Tuple[bool, str]:
        """
        Checks if database schema matches current models and migrations.

        Returns:
            Tuple[bool, str]: (needs_upgrade, status_message)
        """
        try:
            current_rev = self.get_current_revision()
            head_rev = self.get_head_revision()

            if current_rev != head_rev:
                return True, f"Database needs upgrade: {current_rev} -> {head_rev}"

            differences = self.get_schema_differences()
            if differences:
                changes_desc = "\n".join(str(diff) for diff in differences)
                return True, f"Unmigrated changes detected:\n{changes_desc}"

            return False, "Database schema is up to date"

        except Exception as e:
            logger.error(f"Error checking schema status: {str(e)}")
            return True, f"Error checking schema: {str(e)}"

    def upgrade_schema(self, revision: str = "head") -> bool:
        """
        Upgrades database schema to specified revision.

        Args:
            revision: Target revision (default: "head")

        Returns:
            bool: True if upgrade successful
        """
        try:
            config = self.get_alembic_config()
            command.upgrade(config, revision)
            logger.info(f"Schema upgraded successfully to {revision}")
            return True

        except Exception as e:
            logger.error(f"Schema upgrade failed: {str(e)}")
            return False

    def check_and_upgrade(self) -> Tuple[bool, str]:
        """
        Checks schema status and upgrades if necessary.

        Returns:
            Tuple[bool, str]: (action_taken, status_message)
        """
        needs_upgrade, status = self.check_schema_status()

        if needs_upgrade:
            # Remove the auto_upgrade check since we explicitly called this method
            if self.upgrade_schema():
                return True, "Schema was automatically upgraded"
            else:
                return False, "Automatic schema upgrade failed"

        return False, status

    def generate_revision(self, message: str = "auto") -> Optional[str]:
        """
        Generates new migration revision for current schema changes.

        Args:
            message: Revision message

        Returns:
            str: Revision ID if successful, None otherwise
        """
        try:
            config = self.get_alembic_config()
            command.revision(config, message=message, autogenerate=True)
            return self.get_head_revision()

        except Exception as e:
            logger.error(f"Failed to generate revision: {str(e)}")
            return None

    def get_pending_migrations(self) -> List[str]:
        """
        Gets list of pending migrations that need to be applied.

        Returns:
            List[str]: List of pending migration revision IDs
        """
        config = self.get_alembic_config()
        script = ScriptDirectory.from_config(config)

        current = self.get_current_revision()
        head = self.get_head_revision()

        if current == head:
            return []

        pending = []
        for rev in script.iterate_revisions(current, head):
            pending.append(rev.revision)

        return pending

    def print_status(self) -> None:
        """Prints current migration status information to logger."""
        current = self.get_current_revision()
        head = self.get_head_revision()
        differences = self.get_schema_differences()
        pending = self.get_pending_migrations()

        logger.info("=== Database Schema Status ===")
        logger.info(f"Current revision: {current}")
        logger.info(f"Head revision: {head}")
        logger.info(f"Pending migrations: {len(pending)}")
        for rev in pending:
            logger.info(f"  - {rev}")
        logger.info(f"Unmigrated changes: {len(differences)}")
        for diff in differences:
            logger.info(f"  - {diff}")

    def ensure_schema_up_to_date(self) -> bool:
        """
        Reset migrations and create fresh migration for current schema state.
        """
        try:
            logger.info("Resetting migrations and updating to current schema...")

            # 1. Clear the entire alembic directory
            if self.alembic_dir.exists():
                shutil.rmtree(self.alembic_dir)
                logger.info("Cleared alembic directory")

            # 2. Clear alembic_version table
            with self.engine.connect() as connection:
                connection.execute(text("DROP TABLE IF EXISTS alembic_version"))
                connection.commit()
                logger.info("Reset alembic version")

            # 3. Reinitialize alembic from scratch
            if not self._initialize_alembic():
                logger.error("Failed to reinitialize alembic")
                return False

            # 4. Generate fresh migration from current schema
            revision = self.generate_revision("current_schema")
            if not revision:
                logger.error("Failed to generate new migration")
                return False
            logger.info(f"Generated fresh migration: {revision}")

            # 5. Apply the migration
            if not self.upgrade_schema():
                logger.error("Failed to apply migration")
                return False
            logger.info("Successfully applied migration")

            return True

        except Exception as e:
            logger.error(f"Failed to ensure schema is up to date: {e}")
            return False



import uuid
from typing import Any

from fastapi import APIRouter, HTTPException, Query
from sqlmodel import func, select
import structlog

from mtmai.crud import curd_site
from mtmai.crud.curd_site import get_site_by_id
from mtmai.deps import AsyncSessionDep, CurrentUser, OptionalUserDep
from mtmai.models.blog import Tag, TagListResponse
from mtmai.models.models import (
    Item,
)
from mtmai.models.site import (
    ListSiteHostsResponse,
    ListSiteResponse,
    Site,
    SiteCreateRequest,
    SiteHost,
    SiteHostCreateRequest,
    SiteHostCreateResponse,
    SiteHostDeleteResponse,
    SiteHostUpdateRequest,
    SiteHostUpdateResponse,
    SiteItemPublic,
    SiteUpdateRequest,
)

router = APIRouter()
LOG = structlog.get_logger()


@router.get("/site", response_model=ListSiteResponse)
async def list_sites(session: AsyncSessionDep, current_user: OptionalUserDep) -> Any:
    """
    Retrieve site items.
    """
    count_statement = (
        select(func.count()).select_from(Site).where(Item.owner_id == current_user.id)
    )
    a = await session.exec(count_statement)
    count = a.one()
    statement = (
        select(Site).where(
            Site.owner_id == current_user.id
        )  # .offset(skip).limit(limit)
    )
    r = await session.exec(statement)
    items = r.all()

    return ListSiteResponse(data=items, count=count)


@router.get("/site/{id}", response_model=SiteItemPublic)
async def get_site(session: AsyncSessionDep, current_user: CurrentUser, id: str) -> Any:
    """
    Get site by ID.
    """
    site = await get_site_by_id(session, id)
    if site is None or site.owner_id != current_user.id:
        raise HTTPException(status_code=404, detail="not found")
    return site


@router.post("/site/create", response_model=SiteItemPublic)
async def site_create(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    item_in: SiteCreateRequest,
) -> Any:
    return await curd_site.create(
        session, SiteCreateRequest.model_validate(item_in), user_id=current_user.id
    )


@router.put("/site/update/{id}", response_model=SiteItemPublic)
async def site_update(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    id: str,
    item_in: SiteUpdateRequest,
) -> Any:
    """
    Update an item.
    """
    site = await get_site_by_id(session, id)
    if site is None or site.owner_id != current_user.id:
        raise HTTPException(status_code=404, detail="Item not found")

    update_dict = item_in.model_dump(exclude_unset=True)
    site.sqlmodel_update(update_dict)
    session.add(site)
    await session.commit()
    await session.refresh(site)
    return site


@router.get("/hosts", response_model=ListSiteHostsResponse)
async def list_site_hosts(
    session: AsyncSessionDep,
    current_user: OptionalUserDep,
    q: str | None = None,
    siteId: uuid.UUID | None = None,
    skip: int = 0,
    limit: int = 100,
) -> Any:
    """
    Retrieve site hosts.
    """
    count_statement = (
        select(func.count()).select_from(SiteHost).where(SiteHost.site_id == siteId)
    )
    a = await session.exec(count_statement)
    count = a.one()
    statement = (
        select(SiteHost).where(SiteHost.site_id == siteId)  # .offset(skip).limit(limit)
    )
    r = await session.exec(statement)
    items = r.all()

    return ListSiteHostsResponse(data=items, count=count)


@router.post("/hosts", response_model=SiteHostCreateResponse)
async def create_Site_host(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    item_in: SiteHostCreateRequest,
) -> Any:
    """
    Create new site host.
    """
    site_host = SiteHost.model_validate(item_in)
    session.add(site_host)
    await session.commit()
    await session.refresh(site_host)
    return SiteHostCreateResponse(id=site_host.id)


@router.put("/hosts", response_model=SiteHostUpdateResponse)
async def update_site_host(
    *,
    session: AsyncSessionDep,
    current_user: CurrentUser,
    item_in: SiteHostUpdateRequest,
) -> Any:
    """
    Update site host.
    """
    item = await session.get(SiteHost, item_in.id)
    if not item:
        raise HTTPException(status_code=404, detail="Item not found")
    if not current_user.is_superuser and (item.owner_id != current_user.id):
        raise HTTPException(status_code=400, detail="Not enough permissions")
    update_dict = item_in.model_dump(exclude_unset=True)
    item.sqlmodel_update(update_dict)
    session.add(item)
    await session.commit()
    await session.refresh(item)
    return SiteHostUpdateResponse(id=item.id)


@router.delete("/hosts/{id}", response_model=SiteHostDeleteResponse)
async def delete_site_host(
    *, session: AsyncSessionDep, current_user: CurrentUser, id: str
) -> Any:
    """
    Delete site host.
    """
    item = await session.get(SiteHost, id)
    if not item:
        raise HTTPException(status_code=404, detail="Item not found")
    if not current_user.is_superuser and (item.owner_id != current_user.id):
        raise HTTPException(status_code=400, detail="Not enough permissions")
    await session.delete(item)
    await session.commit()
    return SiteHostDeleteResponse(id=item.id)


# tags 相关
@router.get("/tags/list", response_model=TagListResponse)
async def tab_list(
    *,
    session: AsyncSessionDep,
    siteId: uuid.UUID | None = None,
    query: str = Query(default=""),
    offset: int = 0,
    limit: int = Query(default=100, le=100),
):
    count_statement = select(func.count()).select_from(Tag).where(Tag.site_id == siteId)
    a = await session.exec(count_statement)
    count = a.one()
    statement = select(Tag).where(Tag.site_id == siteId).offset(offset).limit(limit)
    r = await session.exec(statement)
    items = r.all()

    return TagListResponse(data=items, count=count)



from fastapi import APIRouter, FastAPI
from loguru import logger


def mount_api_routes(app: FastAPI, prefix=""):
    api_router = APIRouter()

    from mtmai.api import auth

    api_router.include_router(auth.router, tags=["auth"])
    # logger.info("api users")

    # from mtmai.api import users

    # api_router.include_router(users.router, prefix="/users", tags=["users"])

    logger.info("api chat")
    from mtmai.api import chat

    api_router.include_router(chat.router, tags=["chat"])

    # logger.info("api blog")
    # from mtmai.api import blog

    # api_router.include_router(blog.router, prefix="/posts", tags=["posts"])

    # logger.info("api image")
    # from mtmai.api import image

    # api_router.include_router(image.router, prefix="/image", tags=["image"])

    # logger.info("api train")
    # from mtmai.api import train

    # api_router.include_router(train.router, prefix="/train", tags=["train"])

    # logger.info("api metrics")
    # from mtmai.api import metrics

    # api_router.include_router(metrics.router, prefix="/metrics", tags=["metrics"])

    logger.info("api agent")
    from mtmai.api import agent

    api_router.include_router(agent.router, prefix="/agent", tags=["agent"])

    # logger.info("api predict")
    # from mtmai.api import predict

    # api_router.include_router(predict.router, prefix="/predict", tags=["predict"])

    logger.info("api sessions")
    from mtmai.api import sessions

    api_router.include_router(sessions.router, prefix="/sessions", tags=["sessions"])

    # logger.info("api agents")
    # from mtmai.api import agents

    # api_router.include_router(agents.router, prefix="/agents", tags=["agents"])

    # logger.info("api gallery")
    # from mtmai.api import gallery

    # api_router.include_router(gallery.router, prefix="/gallery", tags=["gallery"])

    logger.info("api teams")
    from mtmai.api import teams

    api_router.include_router(teams.router, prefix="/teams", tags=["teams"])

    logger.info("api runs")
    from mtmai.api import runs

    api_router.include_router(runs.router, prefix="/runs", tags=["runs"])

    logger.info("api ws")
    from mtmai.api import ws

    api_router.include_router(ws.router, prefix="/ws", tags=["ws"])

    # LOG.info("api form")
    # from mtmai.api import form

    # api_router.include_router(form.router, prefix="/form", tags=["form"])

    # LOG.info("api site")
    # from mtmai.api import site

    # api_router.include_router(site.router, prefix="/site", tags=["site"])

    # LOG.info("api webpage")
    # from mtmai.api import webpage

    # api_router.include_router(webpage.router, prefix="/webpage", tags=["webpage"])

    # LOG.info("api tasks")
    # from mtmai.api import tasks

    # api_router.include_router(tasks.router, prefix="/tasks", tags=["tasks"])

    logger.info("api openai")
    from mtmai.api import openai

    api_router.include_router(openai.router, tags=["openai"])

    # LOG.info("api listview")
    # from mtmai.api import listview

    # api_router.include_router(listview.router, prefix="/listview", tags=["listview"])

    # LOG.info("api workbench")
    # from mtmai.api import workbench

    # api_router.include_router(workbench.router, prefix="/workbench", tags=["workbench"])

    # logger.info("api logs")
    # from mtmai.api import logs

    # api_router.include_router(logs.router, prefix="/logs", tags=["logs"])

    # LOG.info("api thread")
    # from mtmai.api import thread

    # api_router.include_router(thread.router, prefix="/thread", tags=["thread"])

    # LOG.info("api artifact")
    # from mtmai.api import artifact

    # api_router.include_router(artifact.router, prefix="/artifact", tags=["artifact"])

    logger.info("api config")
    from mtmai.api import config

    api_router.include_router(config.router, prefix="/config", tags=["config"])

    app.include_router(api_router, prefix=prefix)



import uuid

import structlog
from langchain_core.runnables import RunnableConfig
from langgraph.graph import StateGraph
from langgraph.graph.graph import CompiledGraph
from langgraph.prebuilt import tools_condition

from mtmai.agents.ctx import init_mtmai_http_context, mtmai_context
from mtmai.agents.graphutils import (
    create_tool_node_with_fallback,
    is_internal_node,
    is_skip_kind,
    pop_dialog_state,
)
from mtmai.agents.task_graph.nodes.article_gen_node import ArticleGenNode
from mtmai.agents.task_graph.nodes.assisant_node import (
    PrimaryAssistantNode,
    primary_assistant_tools,
    route_assistant,
)
from mtmai.agents.task_graph.nodes.create_task_node import CreateTaskNode
from mtmai.agents.task_graph.nodes.human_input_node import HumanInputNode
from mtmai.agents.task_graph.nodes.site_node import SiteNode
from mtmai.agents.task_graph.nodes.task_entry_node import TaskEntryNode, routeEntryPath
from mtmai.agents.task_graph.task_state import TaskState

LOG = structlog.get_logger()

HUMEN_INPUT_NODE = "human_input"


class TaskGraph:
    def __init__(self):
        pass

    @classmethod
    def name(cls):
        return "blog_graph"

    async def build_graph(self):
        wf = StateGraph(TaskState)

        wf.add_node("entry", TaskEntryNode())
        wf.set_entry_point("entry")
        wf.add_conditional_edges(
            "entry",
            routeEntryPath,
            [
                "articleGen",
                HUMEN_INPUT_NODE,
                "assistant",
                "site",
                "create_task",
            ],
        )

        wf.add_node("assistant", PrimaryAssistantNode())

        wf.add_conditional_edges(
            "assistant",
            tools_condition,
        )

        wf.add_node(
            "tools",
            create_tool_node_with_fallback(primary_assistant_tools),
        )
        wf.add_conditional_edges(
            "tools",
            route_assistant,
            {
                "assistant": "assistant",
                # "error": END,
            },
        )
        wf.add_node(HUMEN_INPUT_NODE, HumanInputNode())
        wf.add_edge(HUMEN_INPUT_NODE, "assistant")

        wf.add_node("articleGen", ArticleGenNode())
        wf.add_edge("articleGen", HUMEN_INPUT_NODE)

        wf.add_node("leave_skill", pop_dialog_state)
        wf.add_edge("leave_skill", "assistant")

        wf.add_node("site", SiteNode())
        wf.add_edge("site", "assistant")

        wf.add_node("create_task", CreateTaskNode())
        # wf.add_edge("create_task", "assistant")

        return wf

    async def on_chat_resume(self):
        init_mtmai_http_context()
        # thread_id = context.session.thread_id
        # user_session = cl.user_session
        # thread_id = context.session.thread_id
        # await cl.Message(content="正在恢复对话").send()
        graph = await TaskGraph().compile_graph()
        # user_session.set("graph", graph)

        # context.session.has_first_interaction = True
        # if not graph:
        #     cl.Message(content="工作流初始化失败").send()
        #     raise ValueError("graph 未初始化")

        thread: RunnableConfig = {
            "configurable": {
                # "thread_id": thread_id,
            }
        }
        pre_state = await graph.aget_state(thread, subgraphs=True)
        # LOG.info("()on_chat_resume %s", thread_id)

        # all_steps = await curd_chat.get_steps_by_thread(thread_id)
        # LOG.info("all_steps %s", all_steps)

    async def onRequest(self, inputs=None, messages=None, thread_id=None):
        graph = await TaskGraph().compile_graph()
        if not graph:
            raise ValueError("graph 未初始化")

        if not thread_id:
            thread_id = str(uuid.uuid4())

        thread: RunnableConfig = {
            "configurable": {
                "thread_id": thread_id,
                # 可以从指定检测点运行，以及分支
                # "checkpoint_id": "xxxxx"
            }
        }
        await self.run_graph(thread, graph, inputs, messages)

    async def run_graph(
        self, thread: RunnableConfig, graph: CompiledGraph, inputs=None, messages=None
    ):
        async for event in graph.astream_events(
            inputs,
            version="v2",
            config=thread,
            subgraphs=True,
        ):
            kind = event["event"]
            node_name = event["name"]
            data = event["data"]

            if not is_internal_node(node_name):
                if not is_skip_kind(kind):
                    LOG.info("[event] %s@%s", kind, node_name)
                    mtmai_context.emit("logs", {"on": kind, "node_name": node_name})

            if kind == "on_chat_model_stream":
                yield data

            if kind == "on_chain_start":
                LOG.info("on_chain_start %s:", node_name)
                output = data.get("output")
                if node_name == "__start__":
                    pass

            if kind == "on_chain_end":
                LOG.info("on_chain_end %s:", node_name)
                output = data.get("output")
                if node_name == "__start__":
                    pass
                if node_name in [HUMEN_INPUT_NODE, "articleGen", "entry"]:
                    human_ouput_message = output.get("human_ouput_message")
                    LOG.info("human_ouput_message %s", human_ouput_message)
            if node_name == "on_chat_start_node":
                thread_ui_state = output.get("thread_ui_state")

            if kind == "on_tool_start":
                pass

            if kind == "on_tool_end":
                output = data.get("output")
                pass
